/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "fsck.h"


//
//
//  check S+ tree of the file system
//
// check_fs_tree stops and recommends to run fsck --rebuild-tree when:
// 1. read fails
// 2. node of wrong level found in the tree
// 3. something in the tree points to wrong block number
//      out of filesystem boundary is pointed by tree
//      to block marked as free in bitmap
//      the same block is pointed from more than one place
//      not data blocks (journal area, super block, bitmaps)
// 4. bad formatted node found
// 5. delimiting keys are incorrect
//      



/* will compare with what does super_block say */
int used_blocks = 0;


/* 1 if block is not marked as used in the on-disk bitmap */
static int is_block_free (struct super_block * s, blocknr_t block)
{
    int i, j;
    char * bitmap;

    i = block / (s->s_blocksize * 8);
    j = block % (s->s_blocksize * 8);
    bitmap = SB_AP_BITMAP (s)[i]->b_data;

    return !test_bit (j, bitmap);
}


/* mark that we have seen this block in the tree already */
static void we_met_it (reiserfs_filsys_t fs, blocknr_t block)
{
    reiserfs_bitmap_set_bit (check_bitmap (fs), block);
    used_blocks (fs) ++;
    return;
}

/* have we seen this block somewhere in the tree before? */
int did_we_meet_it (reiserfs_filsys_t fs, blocknr_t block)
{
    return reiserfs_bitmap_test_bit ((check_bitmap (fs), block);
}


/* FIXME: super block's block count is not checked */
static void init_control_bitmap (reiserfs_filsys_t fs)
{
    int i;

    check_bitmap (fs) = reiserfs_create_bitmap (SB_BLOCK_COUNT (fs));
    if (!check_bitmap (fs)) {
	fprintf (stderr, "init_control_bitmap: could not create bitmap\n");
	return;
    }

    /* skipped and super block */
    for (i = 0; i <= SB_BUFFER_WITH_SB (fs)->b_blocknr; i ++)
        we_met_it (fs, i);

    /* bitmaps */
    for (i = 0; i < SB_BMAP_NR (fs); i ++)
        we_met_it (fs, SB_AP_BITMAP (fs)[i]->b_blocknr);

    /* journal */
    for (i = 0; i < rs_journal_size (fs->s_rs) + 1; i ++)
        we_met_it (fs, i + SB_JOURNAL_BLOCK (fs));
}

/*
void print_bmap_block (int i, char * data, int size, int silent)
{
    int j, k;
    int bits = size * 8;
    int zeros = 0, ones = 0;
  
    printf ("#%d: ", i);

    if (test_bit (0, data)) {
//	 first block addressed by this bitmap block is used
	ones ++;
	if (!silent)
	    printf ("Busy (%d-", i * bits);
	for (j = 1; j < bits; j ++) {
	    while (test_bit (j, data)) {
		ones ++;
		if (j == bits - 1) {
		    if (!silent)
			printf ("%d)\n", j + i * bits);
		    goto end;
		}
		j++;
	    }
	    if (!silent)
		printf ("%d) Free(%d-", j - 1 + i * bits, j + i * bits);

	    while (!test_bit (j, data)) {
		zeros ++;
		if (j == bits - 1) {
		    if (!silent)
			printf ("%d)\n", j + i * bits);
		    goto end;
		}
		j++;
	    }
	    if (!silent)
		printf ("%d) Busy(%d-", j - 1 + i * bits, j + i * bits);

	    j --;
	end:
	}
    } else {
//	 first block addressed by this bitmap is free
	zeros ++;
	if (!silent)
	    printf ("Free (%d-", i * bits);
	for (j = 1; j < bits; j ++) {
	    k = 0;
	    while (!test_bit (j, data)) {
		k ++;
		if (j == bits - 1) {
		    if (!silent)
			printf ("%d)\n", j + i * bits);
		    zeros += k;
		    goto end2;
		}
		j++;
	    }
	    zeros += k;
	    if (!silent)
		printf ("%d) Busy(%d-", j - 1 + i * bits, j + i * bits);
	    
	    k = 0;
	    while (test_bit (j, data)) {
		ones ++;
		if (j == bits - 1) {
		    if (!silent)
			printf ("%d)\n", j + i * bits);
		    ones += k;
		    goto end2;
		}
		j++;
	    }
	    ones += k;
	    if (!silent)
		printf ("%d) Busy(%d-", j - 1 + i * bits, j + i * bits);
	
	    j --;
	end2:
	}
    }

    printf ("used %d, free %d\n", ones, zeros);
}
*/

/* show the difference between 2 bitmaps: disk and control */
/* n - bimap number */
/* disk - pointer to one map */
/* control - pointer to the other */
/* bits - number of blocks in one bitmap */
static void show_diff (int n, char * disk, char * control, int bits)
{
    int i;
    int last_diff = 0;
    int from, num;
    
    from = 0;
    num = 0;
    for (i = 0; i < bits; i ++) {
	if (test_bit (i, disk) && !test_bit (i, control)) {
	    if (last_diff == 1) {
		num ++;
		continue;
	    } else if (last_diff == 2) {
		printf ("Block [%d-%d] free in disk bitmap, used in control\n", from, from + num - 1);
	    }
	    num = 1;
	    from = n * bits + i;
	    last_diff = 1;
	    continue;
	}
	if (!test_bit (i, disk) && test_bit (i, control)) {
	    if (last_diff == 2) {
		num ++;
		continue;
	    } else if (last_diff == 1) {
		printf ("Block [%d-%d] used in disk bitmap, free in control\n", from, from + num - 1);
	    }
	    num = 1;
	    from = n * bits + i;
	    last_diff = 2;
	    continue;
	}
	/* the same bits */
	if (last_diff == 1)
	    printf ("Block [%d-%d] used in disk bitmap, free in control\n", from, from + num - 1);
	if (last_diff == 2)
	    printf ("Block [%d-%d] free in disk bitmap, used in control\n", from, from + num - 1);
	    
	num = 0;
	from = 0;
	last_diff = 0;
	continue;
    }
}
/* this read all blocks which are marked in on-disk and are not marked in control
   bitmap and if read fails mark such blocks in control.
   Return 1 if memcmp is 1 */
static int check_bitmap_for_bad_blocks(struct super_block * s, int bitmap_n, char * disk, char * control, int bytes)
{
  int i, j;
  struct buffer_head * bh;

  for (i = 0; i < bytes; i ++)
    if (disk[i] != control[i])
      for (j = 0; j<8; j++)
        if (test_bit(j, &disk[i]) && !test_bit(j, &control[i])) {
          if ((bh = bread (s->s_dev, bitmap_n * bytes * 8 + i * 8 + j, s->s_blocksize)) != 0)
            brelse(bh);
          else
            we_met_it (s, bitmap_n * bytes * 8 + i * 8 + j);
        }
  return (memcmp(disk, control, bytes));
}

/* compare on-disk and control(g_new_bitmap for FAST-REBUILD) bitmap */
/* show the difference */
/* correct bitmap for CORRECT-BITMAP mode */
static void compare_bitmaps (struct super_block * s)
{
    int i, wrong_bitmap = 0;
    char * bitmap;

    printf ("Comparing bitmaps..");

    if (SB_FREE_BLOCKS (s) != SB_BLOCK_COUNT (s) - used_blocks) {
	printf ("\nUsed blocks %d, super block version %d - %d", used_blocks, SB_BLOCK_COUNT (s), SB_FREE_BLOCKS (s));
	wrong_bitmap = 1;
    }

    fflush(stdout);
    for (i = 0; i < SB_BMAP_NR (s); i ++) {
	if (opt_fsck_mode == FSCK_DEFAULT || opt_fsck_mode == FSCK_CORRECT_BITMAP)
        {
        /* we are read-only checking the partition, check this bitmap */
          bitmap = control_bitmap[i];
        }else
          /* we are re-building the tree, check this bitmap */
          bitmap = g_new_bitmap [i];

	if (memcmp (SB_AP_BITMAP(s)[i]->b_data, bitmap, s->s_blocksize))
        /* check_bitmap_for_bad_blocks, mark bad blocks as used in control_bitmap
           and compare bitmaps (memcmp) after that */
        if (check_bitmap_for_bad_blocks(s, i, SB_AP_BITMAP(s)[i]->b_data, bitmap, s->s_blocksize))
        {
          printf ("\nbitmap %d does not match to the correct one", i);
	    if (opt_verbose) {
      	printf ("\nSee diff");
            show_diff (i, SB_AP_BITMAP(s)[i]->b_data, bitmap, s->s_blocksize * 8);
          }
    	    if (opt_fsck_mode == FSCK_CORRECT_BITMAP)
          {
            memcpy(SB_AP_BITMAP(s)[i]->b_data, bitmap, s->s_blocksize);
            mark_buffer_dirty(SB_AP_BITMAP(s)[i], 1);
//        	printf("Fixed changed bitmap block: %d ", SB_AP_BITMAP(s)[i]->b_blocknr);
            printf(" Fixed ");
          }
          wrong_bitmap = 1;
        }
    }
    if (wrong_bitmap)
    {
        if (opt_fsck_mode == FSCK_CORRECT_BITMAP)
            printf ("\n");
        else
            reiserfs_panic (s, "\nRun reiserfsck with --rebuild-tree (or rewrite correct bitmap)\n");
    }
    printf ("ok\n");
}

/* is this block legal to be pointed to by some place of the tree? */
static int bad_block_number (struct super_block * s, blocknr_t block)
{
    if (block >= SB_BLOCK_COUNT (s)) {
        reiserfs_warning ("block out of filesystem boundary found\n");
        return 1;
    }

    if (not_data_block (s, block)) {
        reiserfs_warning ("not data block is used in the tree\n");
        return 1;
    }

    if ((opt_fsck_mode == FSCK_DEFAULT) && is_block_free (s, block)) {
        reiserfs_warning ("block %lu is not marked as used in the disk bitmap\n", block);
        return 1;
    }

    return 0;
}

/*
1 if the block is used already in new bitmap
	(control for CHECKING - was_block_used)
	(g_new_bitmap for REBUILDING - is_block_used)
0 otherwise and and mark the block as used for DEFAULT and CORRECT_BITMAP
*/
static int got_already(struct super_block * s, blocknr_t block)
{
    if (opt_fsck_mode == FSCK_FAST_REBUILD){
	if (is_block_used(block)){
	  reiserfs_warning ("block %lu is in tree already\n", block);
	  return 1;
	}
    }else{
	if (did_we_meet_it (s, block)) {
	  reiserfs_warning ("block %lu is in tree already\n", block);
	  return 1;
	}
	we_met_it (s, block);
    }
    return 0;
}

/* 1 if some of fields in the block head of bh look bad */
int bad_block_head (struct buffer_head * bh)
{
    struct block_head * blkh;

    blkh = B_BLK_HEAD (bh);
    if (le16_to_cpu (blkh->blk_nr_item) > (bh->b_size - BLKH_SIZE) / IH_SIZE) {
//	reiserfs_warning ("block %lu has wrong blk_nr_items (%z)\n",  bh->b_blocknr, bh);
	return 1;
    }
    if (le16_to_cpu (blkh->blk_free_space) > 
	bh->b_size - BLKH_SIZE - IH_SIZE * le16_to_cpu (blkh->blk_nr_item)) {
//	reiserfs_warning ("block %lu has wrong blk_free_space %z\n", bh->b_blocknr, bh);
	return 1;
    }
    return 0;
}


/* 1 if it does not look like reasonable stat data */
static int bad_stat_data (struct buffer_head * bh, struct item_head * ih)
{
#ifndef __alpha__
    if (opt_fsck_mode != FSCK_FAST_REBUILD)
    {
/*        if (!is_objectid_used (&g_sb, ih->ih_key.k_objectid)) {
            reiserfs_warning ("%lu is marked free, but used by an object", ih->ih_key.k_objectid);
            return 1;
        }*/

        // FIXME: this could be cured right here
        if (mark_id_as_used(ih->ih_key.k_objectid))
            reiserfs_warning("\n\nbad_stat_data: objectid was used already %k", &ih->ih_key);
    }
#endif
    return 0;
}


/* it looks like we can check item length only */
static int bad_direct_item (struct buffer_head * bh, struct item_head * ih)
{
    return 0;
}


/* each unformatted node pointer*/
static int bad_indirect_item (struct super_block * s, struct buffer_head * bh,
			      struct item_head * ih)
{
    int i;
    __u32 * ind = (__u32 *)B_I_PITEM (bh, ih);

    if (ih_item_len(ih) % 4)
	return 1;

    for (i = 0; i < I_UNFM_NUM (ih); i ++) {
	/* check unformatted node pointer and mark it used in the control bitmap */
        if (ind[i] && bad_block_number (s, le32_to_cpu (ind[i])))
	        return 1;
        if (ind[i] && got_already(s, le32_to_cpu (ind[i])))
            return 1;
    }

    /* delete this check for 3.6 */
    if (ih_free_space (ih) > s->s_blocksize - 1)
	reiserfs_warning ("%h has wrong ih_free_space\n", ih);
    return 0;
}


/* check entry count and locations of all names */
static int bad_directory_item (struct buffer_head * bh, struct item_head * ih)
{
    int i;
    struct reiserfs_de_head * deh;


    if (ih_entry_count (ih) > ih_item_len (ih) / (DEH_SIZE + 1))
	return 1;

    deh = B_I_DEH (bh, ih);
    for (i = 0; i < ih_entry_count (ih); i ++, deh ++) {
	if (deh_location(deh) >= ih_item_len(ih))
	    return 1;
	if (i && deh_location(deh) >= deh_location(deh-1))
	    return 1;
	if ((ih->ih_key.k_objectid != REISERFS_ROOT_OBJECTID && deh_dir_id (deh) == 0) ||
	    deh_offset (deh) == 0 || deh_objectid (deh) == 0 || 
	    deh_dir_id (deh) == deh_objectid (deh))
	    return 1;
    }
    return 0;
}


static int bad_item (struct super_block * s, struct buffer_head * bh, int i)
{
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, i);
    if (is_stat_data_ih (ih))
	return bad_stat_data (bh, ih);

    if (is_direct_ih (ih))
	return bad_direct_item (bh, ih);

    if (is_indirect_ih(ih))
	return bad_indirect_item (s, bh, ih);
    
    return bad_directory_item (bh, ih);
}


/* 1 if i-th and (i-1)-th items can not be neighbors */
/* For pass1a and pass2 we should control increasing of key only.
   For FSCK_DAFAULT or pass1 (we havn't changed content of items)
   we can check types of neighbouring items also.
 */
int bad_pair (struct super_block * s, struct buffer_head * bh, int i)
{
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, i);

    if (comp_keys (&((ih - 1)->ih_key), &ih->ih_key) != -1)
	return 1;

    if (is_stat_data_ih (ih))
	/* left item must be of another object */
	if (comp_short_keys (&((ih - 1)->ih_key), &ih->ih_key) != -1)
	    return 1;

    /* opt_fsck_mode:	FSCK_REBUILD, FAST_REBUILD, FSCK_DEFAULT, FSCK_CORRECT_BITMAPS
       fsck_stage:	PASS1a or PASS2 */
    if (opt_fsck_mode != FSCK_DEFAULT && opt_fsck_mode != FSCK_CORRECT_BITMAP && fsck_stage != PASS1)
      return 0;

    if (is_direct_ih(ih)) {
	/* left item must be indirect or stat data item of the same file for
           --check */
	/* left item must be indirect or stat data item if only it is the same
           file --rebuild */
	if (opt_fsck_mode == FSCK_DEFAULT) {
            if (not_of_one_file (&((ih - 1)->ih_key), &ih->ih_key))
                return 1;

            if (!((is_stat_data_ih (ih - 1) && get_offset (&ih->ih_key) == 1) ||
                   (is_indirect_ih (ih - 1) &&	
                    get_offset (&(ih - 1)->ih_key) + get_bytes_number (ih-1, bh->b_size) ==
                    get_offset (&ih->ih_key))))
                return 1;
        }else if (!not_of_one_file (&((ih - 1)->ih_key), &ih->ih_key)) {
            if (!((is_stat_data_ih (ih - 1) && get_offset (&ih->ih_key) == 1) ||
                   (is_indirect_ih (ih - 1) &&	
                    get_offset (&(ih - 1)->ih_key) + get_bytes_number (ih-1, bh->b_size) ==
                    get_offset (&ih->ih_key))))
                return 1;
        }
    }

    if (is_indirect_ih (ih) || is_direntry_ih (ih)) {
	/* left item must be stat data of the same object */
      if (not_of_one_file (&((ih - 1)->ih_key), &ih->ih_key))
        return 1;

      if (!is_stat_data_ih (ih - 1))
        return 1;
    }
    return 0;
}

/* 1 if block head or any of items is bad or
   an item cannot have such a neighbouring item */
static int bad_leaf (struct super_block * s, struct buffer_head * bh)
{
    int i;

    if (bad_block_head (bh))
	return 1;
    
    for (i = 0; i < B_NR_ITEMS (bh); i ++) {
	if (bad_item (s, bh, i)) {
	    reiserfs_warning ("block %lu has invalid item %d: %h\n", bh->b_blocknr, i, B_N_PITEM_HEAD (bh, i));
	    return 1;
	}
	
	if (i && bad_pair (s, bh, i)) {
	    reiserfs_warning ("block %lu has wrong order of items\n", bh->b_blocknr);
	    return 1;
	}
    }
    return 0;
}


static inline int is_key_left_mergeable (struct key * key, unsigned long bsize)
{
    if (KEY_IS_DIRECT_KEY (key))
	return ((get_offset (key) & (bsize - 1)) != 1);

    if (KEY_IS_INDIRECT_KEY (key))
	return (get_offset (key) != 1);
    
    if (KEY_IS_DIRECTORY_KEY (key))
	return (key->u.k_offset_v1.k_offset != DOT_OFFSET);
    
#ifdef CONFIG_REISERFS_CHECK
    if ( ! KEY_IS_STAT_DATA_KEY (key))
	reiserfs_panic (0, "vs-16060: is_key_left_mergeable: item [%k] must be a stat data", key);
#endif

    return 0;
}


/* 1 if bh does not look like internal node */
static int bad_internal (struct super_block * s, struct buffer_head * bh)
{
    int i;
    int left_child_size, right_child_size;
    
    for (i = 0; i <= B_NR_ITEMS (bh); i ++) {
        if (i < B_NR_ITEMS (bh) - 1)
            if (comp_keys (B_N_PDELIM_KEY (bh, i), 
			   B_N_PDELIM_KEY (bh, i + 1)) != -1)
		/* 'i'-th key is not less than 'i + 1'-th one */
                return 1;
        if (bad_block_number (s, child_block_number (bh, i))){
            return 1;
        }

	if (!i)
	    continue;

	// make sure that balance condition is not broken
	left_child_size = child_block_size (bh, i - 1);
	right_child_size = child_block_size (bh, i);

	if (node_level (bh) == DISK_LEAF_NODE_LEVEL + 1) {
	    if (left_child_size + right_child_size -
		(is_key_left_mergeable (B_N_PDELIM_KEY (bh, i - 1), bh->b_size) ? IH_SIZE : 0) <=
		MAX_CHILD_SIZE (bh)) {
		reiserfs_warning ("vs-: bad_internal: balance condition is "
				  "broken between %d and %d children (leaves) of block %d\n",
				  i - 1, i, bh->b_blocknr);
	    }
	} else {
	    if (left_child_size + right_child_size + KEY_SIZE <= 
		MAX_CHILD_SIZE (bh)) {
		reiserfs_warning ("vs-: bad_internal: balance condition is "
				  "bloken between %d and %d children (internals) of block %d\n",
				  i - 1, i, bh->b_blocknr);
	    }
	}
    }
    return 0;
}

/* h == 0 for root level. block head's level == 1 for leaf level  */
static inline int h_to_level (struct super_block * s, int h)
{
    return SB_TREE_HEIGHT (s) - h - 1;
}

/* bh must be formatted node. blk_level must be tree_height - h + 1 */
static int bad_node (struct super_block * s, struct buffer_head ** path, int h)
{
    struct buffer_head ** pbh = &path[h];

    if (B_LEVEL (*pbh) != h_to_level (s, h)) {
	reiserfs_warning ("node with wrong level found in the tree %d-%d\n", B_LEVEL (*pbh), h_to_level (s, h));
	return 1;
    }

    if (bad_block_number (s, (*pbh)->b_blocknr))
	return 1;
    if (got_already(s, (*pbh)->b_blocknr))
	return 1;

    if (is_leaf_node (*pbh))
	return bad_leaf (s, *pbh);

    return bad_internal (s, *pbh);
}


/* return the position in internal node which contains
   the pointer to this block */
static int get_pos (struct buffer_head * bh, blocknr_t block)
{
    int i;

    for (i = 0; i <= B_NR_ITEMS (bh); i ++) {
	if (child_block_number (bh, i) == block)
	    return i;
    }
    die ("get_pos: position for block %lu not found", block);
    return 0;
}


/* path[h] - leaf node */
static struct key * lkey (struct buffer_head ** path, int h)
{
    int pos;

    while (h > 0) {
       pos = get_pos (path[h - 1], path[h]->b_blocknr);
       if (pos)
           return B_N_PDELIM_KEY(path[h - 1], pos - 1);
       h --;
    }
    return 0;
}


/* path[h] - leaf node */
static struct key * rkey (struct buffer_head ** path, int h)
{
    int pos;

    while (h > 0) {
       pos = get_pos (path[h - 1], path[h]->b_blocknr);
       if (pos != B_NR_ITEMS (path[h - 1]))
           return B_N_PDELIM_KEY (path[h - 1], pos);
       h --;
    }
    return 0;
}


/* are all delimiting keys correct */
static int bad_path (struct super_block * s, struct buffer_head ** path, int h1)
{
  int h = 0;
  struct key * dk;

  while (path[h])
      h ++;

  h--;

  // path[h] is leaf
  if (h != h1)
      die ("bad_path: wrong path");

  dk = lkey (path, h);
  if (dk && comp_keys (dk, B_N_PKEY (path[h], 0)))
      // left delimiting key must be equal to the key of 0-th item in the node
      return 1;

  dk = rkey (path, h);
  if (dk && comp_keys (dk, B_N_PKEY (path[h], node_item_number (path[h]) - 1)) != 1)
      // right delimiting key must be bigger than the key of the last item in the node
      return 1;

  return 0;
}


/* pass the S+ tree of filesystem */
void check_fs_tree (reiserfs_filsys_t fs)
{
    struct fsck_info info = {0, };

    fs->s_vp = &info;
    init_control_bitmap (fs);

    init_id_map();

    fprintf (stderr, "Checking S+tree..");

    pass_through_tree (s, bad_node, bad_path);

    /* S+ tree is correct (including all objects have correct sequences of items) */
    printf ("ok\n");

    /* compare created bitmap with the original */
    compare_bitmaps (s);

    compare_id_maps(s);
//    free_objectid_maps();
    free_id_map();
}


/* this removes internal pointer from the tree to the last block in the path		*/
/* from the previous block in the path							*/
/* if the internal block is left without any pointer, remove it also			*/
void remove_internal_pointer(struct super_block * s, struct buffer_head ** path)
{
    int h = 0;
    int pos, items;
    __u32 block;

    while (path[h])
        h ++;

    h--;
    block = path[h]->b_blocknr;
        printf("\nremove pointer to (%d) block", block);
    brelse(path[h]);
    path[h] = 0;
    h--;
    while (h>=0)
    {
    	  /* remove whole internal block from the tree if it contain only one pointer */
        if (B_NR_ITEMS(path[h]) <= 1)
        {
            block = path[h]->b_blocknr;
            brelse(path[h]);
            path[h] = 0;
            mark_block_free(block);
            unmark_block_formatted(block);
            used_blocks++;
            h --;
            continue;
        }
        pos = get_pos (path[h], block);
        if (pos)
        {
            memmove (B_N_CHILD(path[h],pos), B_N_CHILD(path[h],pos+1),
                s->s_blocksize - BLKH_SIZE - B_NR_ITEMS(path[h])*KEY_SIZE - DC_SIZE*(pos+1));
            memmove(B_N_PDELIM_KEY(path[h],pos-1), B_N_PDELIM_KEY(path[h],pos),
                s->s_blocksize - BLKH_SIZE - (pos)*KEY_SIZE);
        }else{
            __u32 move_block = path[h]->b_blocknr;
            int move_to_pos;
            int height = h;

		/* upper pointer to the whole right branch must be corrected for leftmost leaves */
            while(--height >= 0)
            {
                move_to_pos = get_pos (path[height], move_block);
                if (move_to_pos == 0){
                    move_block = path[height]->b_blocknr;
                    continue;
                }
                *B_N_PDELIM_KEY(path[height], move_to_pos-1) = *B_N_PDELIM_KEY(path[h], 0);
                break;
            }

            memmove (B_N_CHILD(path[h], 0), B_N_CHILD(path[h], 1),
                s->s_blocksize - BLKH_SIZE - B_NR_ITEMS(path[h])*KEY_SIZE - DC_SIZE);
            memmove(B_N_PDELIM_KEY(path[h], 0), B_N_PDELIM_KEY(path[h], 1),
                s->s_blocksize - BLKH_SIZE - KEY_SIZE);
        }
        set_node_item_number(path[h], node_item_number(path[h]) - 1);
        mark_buffer_dirty(path[h], 1);
        break;
    }

    /* all nodes were removed from the tree */
    if (h == -1)
    {
        SB_DISK_SUPER_BLOCK(s)->s_root_block = ~0;
        SB_DISK_SUPER_BLOCK(s)->s_tree_height = ~0;
        mark_buffer_dirty(SB_BUFFER_WITH_SB(s), 1);
    }
}

void handle_buffer(struct super_block * s, struct buffer_head * bh)
{
    int i, j;
    struct item_head * ih;

    if (is_leaf_node (bh))
    {
        for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++)
        {
        	/* look through all indirect items pointers and mark them as used and unformatted */
            if (is_indirect_ih(ih))
                for (j = 0; j < I_UNFM_NUM (ih); j ++)
                    if (B_I_POS_UNFM_POINTER(bh,ih,j)){
                        mark_block_unformatted(le32_to_cpu(B_I_POS_UNFM_POINTER(bh,ih,j)));
                        mark_block_used(le32_to_cpu(B_I_POS_UNFM_POINTER(bh,ih,j)));
                        used_blocks++;
                    }
        	/* handle SD */
        	if (is_stat_data_ih (ih)) {
	            add_event (STAT_DATA_ITEMS);
                if (ih_key_format(ih) == KEY_FORMAT_1)
                    ((struct stat_data_v1 *)B_I_PITEM(bh,ih))->sd_nlink = 0;
                else
                    ((struct stat_data *)B_I_PITEM(bh,ih))->sd_nlink = 0;
                mark_buffer_dirty(bh, 1);
        	}
        }
    }
    mark_block_used(bh->b_blocknr);
//    we_met_it(s, bh->b_blocknr);
    mark_block_formatted(bh->b_blocknr);
    used_blocks++;
}
	
/* bh must be formatted node. blk_level must be tree_height - h + 1 */
static int handle_node (struct super_block * s, struct buffer_head ** path, int h)
{
    if (bad_node(s, path, h)){
       remove_internal_pointer(s, path);
       return 1;
    }
    handle_buffer(s, path[h]);
    return 0;
}

/* are all delimiting keys correct */
static int handle_path (struct super_block * s, struct buffer_head ** path, int h)
{
    if (bad_path(s, path, h)){
        remove_internal_pointer(s, path);
        return 1;
    }
    return 0;
}

/* return 1 if rebuild tree from scratch is needed */
void check_internal_structure(struct super_block * s)
{
    /* control bitmap is used to keep all blocks we should not put into tree again */
    /* used bitmap is used to keep all inserted blocks. The same as control bitmap plus unfm blocks */
//    init_control_bitmap(s);

    printf ("Checking S+tree..");

    pass_through_tree (s, handle_node, handle_path);

//    compare_bitmaps(s);
    printf ("ok\n");
}

int check_sb (struct super_block * s)
{
    int format_sb = 0;
    int problem = 0;
    struct reiserfs_super_block * rs;
    __u32 block_count;

    rs = s->s_rs;
    /* in (REISERFS_DISK_OFFSET_IN_BYTES / 4096) block */
    if (is_reiser2fs_magic_string (rs) &&
        SB_JOURNAL_BLOCK(s) == get_journal_start_must (rs_blocksize (rs)))
    {
	/* 3.6 or >=3.5.22 */
	printf("\t  3.6.x format SB found\n");
        format_sb = 1;
        goto good_format;
    }

    if (is_reiserfs_magic_string (rs) &&
        SB_JOURNAL_BLOCK(s) == get_journal_start_must (rs_blocksize (rs)))
    {
	/* >3.5.9(10) and <=3.5.21 */
        printf("\t>=3.5.9 format SB found\n");
        format_sb = 2;
        goto good_format;
    }

    /* in 2 block */
    if (is_reiser2fs_magic_string (rs) &&
        SB_JOURNAL_BLOCK(s) == get_journal_old_start_must (rs))
    {
	/* <3.5.9(10) converted to new format */
        printf("\t< 3.5.9(10) SB converted to new format found \n");
        format_sb = 3;
        goto good_format;
    }
	
    if (is_reiserfs_magic_string (rs) &&
        SB_JOURNAL_BLOCK(s) == get_journal_old_start_must (rs))
    {
	/* <3.5.9(10) */
        printf("\t< 3.5.9(10) format SB found\n");
        format_sb = 4;
        goto good_format;
    }
    else	
	die("check SB: wrong SB format found\n");
	
good_format:	
	
    if (s->s_blocksize != 4096) {
	reiserfs_warning("check SB: specified block size (%d) is not correct must be 4096\n", s->s_blocksize);
	problem++;
    }
    
    /* for 4096 blocksize only */
    if ((rs_tree_height(rs) < DISK_LEAF_NODE_LEVEL) || (rs_tree_height(rs) > MAX_HEIGHT)){
	reiserfs_warning("check SB: wrong tree height (%d)\n", rs_tree_height(rs));
	problem++;
    }

    block_count = count_blocks ("", s->s_blocksize, s->s_dev);

    if (SB_BLOCK_COUNT(s) > block_count){
	reiserfs_warning("check SB: specified block number (%d) is too high\n", SB_BLOCK_COUNT(s));
	problem++;
    }

    if ((rs_root_block(rs) >= block_count) || (rs_root_block(rs) < 0)){
	reiserfs_warning("check SB: specified root block number (%d) is too high\n", rs_root_block(rs));
	problem++;
    }

    if (SB_FREE_BLOCKS(s) > SB_BLOCK_COUNT(s)){
	reiserfs_warning("check SB: specified free block number (%d) is too high\n", SB_FREE_BLOCKS(s));
	problem++;
    }		
    
    if (SB_REISERFS_STATE(s) != REISERFS_VALID_FS){
	reiserfs_warning("check SB: wrong (%d) state\n", SB_REISERFS_STATE(s));
	problem++;
    }		
    
    if ( SB_BMAP_NR(s) != SB_BLOCK_COUNT(s) / (s->s_blocksize * 8) +
	 ((SB_BLOCK_COUNT(s) % (s->s_blocksize * 8)) ? 1 : 0)){
	reiserfs_warning("check SB: wrong bitmap number (%d)\n", SB_BMAP_NR(s));
	problem++;
    }		

    if (SB_VERSION(s) == REISERFS_VERSION_2 || SB_VERSION(s) == REISERFS_VERSION_1)
    {
	if (!(SB_VERSION(s) == REISERFS_VERSION_2 && (format_sb == 1 || format_sb == 3)) &&
            !(SB_VERSION(s) == REISERFS_VERSION_1 && (format_sb == 2 || format_sb == 4))){
	  reiserfs_warning("check SB: wrong SB version == %d, format == %d\n", SB_VERSION(s), format_sb);
	  problem++;
	}		
    }else{
	reiserfs_warning("check SB: wrong SB version (%d)\n", SB_VERSION(s));
	problem++;
    }		

    if (SB_VERSION(s) == REISERFS_VERSION_2 &&
        (rs_hash (rs) < 1 || rs_hash (rs) > 3)) {
	reiserfs_warning("check SB: wrong hash (%d)\n", rs_hash (rs));
	problem++;
    }		


    if ((SB_VERSION(s) == REISERFS_VERSION_2) ?
	(rs_objectid_map_max_size (rs) != ((s->s_blocksize - SB_SIZE) / sizeof(__u32) / 2 * 2)) :
	(rs_objectid_map_max_size (rs) != ((s->s_blocksize - SB_SIZE_V1) / sizeof(__u32) / 2 * 2)))
    {
	reiserfs_warning("check SB: objectid map corrupted max_size == %d\n", rs_objectid_map_max_size (rs));
	problem++;
    }

    if (rs_objectid_map_size (rs) < 2 ||
	rs_objectid_map_size (rs) > rs_objectid_map_max_size (rs))
    {
	reiserfs_warning("check SB: objectid map corrupted cur_size == %d\n", rs_objectid_map_size (rs));
	problem++;
    }		

    if (rs_journal_size(rs) != JOURNAL_BLOCK_COUNT){
	reiserfs_warning("check SB: specified journal size (%d) is not correct must be %d\n",
		rs_journal_size(rs), JOURNAL_BLOCK_COUNT);
	problem++;
    }


    if (!problem)
	reiserfs_warning("\t  No problem found\n");
	
    return format_sb;
}
