/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "fsck.h"
#include <getopt.h>
#include <sys/mount.h>

#include "../version.h"


/* fsck man 
 there are 3 modes of fsck : 
 1. check
 	this is default mode. Use to check filesystem consistency.
 2. rebuild-tree
 	
 3. rebuild-sb
 4. 
*/

#define print_usage_and_exit() die ("Usage: %s [-aprvy] [--rebuild-tree]\n"\
"[--scan-whole-partition] [--no-journal-replay]\n"\
"[--replay-whole-journal] device\n"\
"\n"\
"Long options:\n"\
"\tModes:\n"\
"\t--check                  consistency checking (default)\n"\
"\t--correct-bitmap         consistency checking with correcting bitmap\n"\
"\t--rebuild-sb             super block checking and rebuilding if needed\n"\
"\t--rebuild-tree           force fsck to rebuild filesystem from scratch\n"\
"\t                         (takes a long time)\n"\
"\t--search-key             will search for key you specify\n"\
"\n"\
"\tArea to scan:\n"\
"\t--scan-used-part-only    scan what is marked used in bitmap (default)\n"\
"\t--scan-whole-partition   scan whole partition\n"\
"\n"\
"\tJournal replay options:\n"\
"\t--replay-by-mount        replay by calling mount -o replay-only (default)\n"\
"\t--no-journal-replay      skip journal replaying\n"\
"\t--replay-whole-journal   replay all valid transaction found in the journal\n"\
"\n"\
"Short options:\n"\
"\t-v verbose mode\n"\
"\t-a supress progress information\n"\
"\t-y\n"\
"\t-p do nothing, exist for compatibility with fsck(8)\n"\
"\t-r\n", argv[0]);





/* fsck is called with one non-optional argument - file name of device
   containing reiserfs. This function parses other options, sets flags
   based on parsing and returns non-optional argument */
static char * parse_options (int argc, char * argv [], struct what_to_do * job)
{
    int c;
    int i;
    struct option options[] = {
      {"check", no_argument, &job->mode, FSCK_CHECK},
      {"rebuild-sb", no_argument, &job->mode, FSCK_MAKE_SB},
      {"rebuild-tree", no_argument, &job->mode, FSCK_REBUILD},

      {"correct-easy-fixable", no_argument, &job->fix_fixable, 1},
      
      {"journal-replay", no_argument, &job->do_journal_replay, 1},
      {"no-journal-replay", no_argument, &job->do_journal_replay, 0},

      {"force", no_argument, &job->force, 1},
      {"Yes", no_argument, &job->yes_confirmed, 1},
      {"verbose", no_argument, &job->verbose, 1},

      /* {"fast-rebuild", no_argument, &opt_fsck_mode, FSCK_FAST_REBUILD}, */
      /* {"scan-whole-partition", no_argument, &opt_what_to_scan, SCAN_WHOLE_PARTITION}, */
      /* {"no-lost+found", no_argument, &opt_lost_found, NO_LOST_FOUND}, */
      /* {"lost+found",  no_argument, &opt_lost_found, DO_LOST_FOUND}, */

      {0, 0, 0, 0}
    };
    int option_index;


    while (1) {
      
	c = getopt_long (argc, argv, "yapv", options, &option_index);
	if (c == -1)
	    break;

	switch (c) {
	case 0:
	    /* long option found */
	    break;

	case 'y':
	case 'p': /* these do nothing */
	    break;

	case 'a':
	    opt_fsck = 1;
	    break;

	case 'f':
	    job->force = 1;
	    break;

	case 'v':
	    /* print some information on run and on exit */
	    job->verbose = 1;
	    break;

	default:
	    print_usage_and_exit();
	}
    }
    
    if (optind != argc - 1)
	/* only one non-option argument is permitted */
	print_usage_and_exit();
    
    return argv[optind];
}


#if 0
struct super_block g_sb;
struct buffer_head * g_sbh;
struct reiserfs_super_block * g_old_rs;
struct reiserfs_transaction_handle g_th;


static void reset_super_block (struct super_block * s)
{
    __u32 * oids;
    struct reiserfs_super_block * rs;

    g_old_rs = (struct reiserfs_super_block *)getmem (s->s_blocksize);
    memcpy (g_old_rs, SB_BUFFER_WITH_SB (s)->b_data, s->s_blocksize);

    rs = (struct reiserfs_super_block *)(SB_BUFFER_WITH_SB (s)->b_data);

    /* reset few fields in */
    set_free_blocks (rs, SB_BLOCK_COUNT (s));
    if (opt_fsck_mode != FSCK_FAST_REBUILD){
        set_root_block (rs, ~0);
        set_tree_height (rs, ~0);
    }
    set_state (rs, REISERFS_ERROR_FS);
//    set_objectid_map_size (rs, 2);


    // reset objectid map
//    oids = (__u32 *)((char *)rs + (sb_size(s)));
//    if (oids[0] != cpu_to_le32 (1)) {
//	printf ("reset_super_block: invalid objectid map\n");
//	oids[0] = cpu_to_le32 (1);
//    }
//    oids[1] = cpu_to_le32 (2);

    // hash functino will be found in semantic pass
    s->s_hash_function = 0;

    mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 0);
    /* make file system invalid unless fsck done () */
    bwrite (SB_BUFFER_WITH_SB (s));

}


static void update_super_block (void)
{
    struct reiserfs_super_block * rs;

    rs = (struct reiserfs_super_block *)(SB_BUFFER_WITH_SB (&g_sb)->b_data);

    set_state (rs, REISERFS_VALID_FS);

    reset_journal (&g_sb);

    mark_buffer_dirty (SB_BUFFER_WITH_SB (&g_sb), 0);
}


char ** g_disk_bitmap;
char ** g_new_bitmap;
char ** g_uninsertable_leaf_bitmap;
char ** g_formatted;
char ** g_unformatted;

__u32 g_blocks_to_read_in_pass0;
__u32 g_blocks_to_read_in_pass1;



/* read bitmaps (new or old format), create data blocks for new
   bitmap, mark non-data blocks in it (skipped, super block, journal
   area, bitmaps) used, create other auxiliary bitmaps */
static void init_bitmaps (struct super_block * s)
{
    int i, j;

    /* read disk bitmap */
    uread_bitmaps (s);

//    if (uread_bitmaps (s))
//      die ("init_bitmap: unable to read bitmap");

    g_disk_bitmap = getmem (sizeof (char *) * SB_BMAP_NR (s));
    for (i = 0; i < SB_BMAP_NR (s); i ++) {
	g_disk_bitmap[i] = SB_AP_BITMAP (s)[i]->b_data;

	if (opt_what_to_scan == SCAN_WHOLE_PARTITION)
	    /* mark all blocks busy */
	    memset (g_disk_bitmap[i], 0xff, s->s_blocksize);
    }

    /* g_blocks_to_read is used to report progress */
    if (opt_what_to_scan == SCAN_WHOLE_PARTITION)
        /* all blocks will be scanned except journal blocks, bitmaps, SB and reserved space */
        g_blocks_to_read_in_pass0 = SB_BLOCK_COUNT (s) - (SB_JOURNAL_BLOCK (s) + JOURNAL_BLOCK_COUNT) - SB_BMAP_NR(&g_sb);
    else {
        /* blocks marked used in bitmap will be scanned */
        g_blocks_to_read_in_pass0 = 0;
        for (i = 0; i < SB_BMAP_NR (s); i ++)
            for (j = 0; j < s->s_blocksize * 8; j ++)
                if (!not_data_block(&g_sb, i * s->s_blocksize * 8 + j) &&
                    i * s->s_blocksize * 8 + j < SB_BLOCK_COUNT (s) &&
                    test_bit (j, SB_AP_BITMAP (s)[i]->b_data))
                    g_blocks_to_read_in_pass0 ++;
    }

    /* this bitmap will contain valid bitmap when fsck will have done */
    g_new_bitmap = getmem (sizeof (char *) * SB_BMAP_NR (s));
    for (i = 0; i < SB_BMAP_NR (s); i ++)
	g_new_bitmap[i] = getmem (s->s_blocksize);

    /* mark skipped blocks and super block used */
    for (i = 0; i <= SB_BUFFER_WITH_SB (s)->b_blocknr; i ++)
	mark_block_used (i);

    /* mark bitmap blocks as used */
    for (i = 0; i < SB_BMAP_NR (s); i ++)
	mark_block_used (SB_AP_BITMAP (s)[i]->b_blocknr);

    /* mark journal area as used */
    for (i = 0; i < JOURNAL_BLOCK_COUNT + 1; i ++)
	mark_block_used (i + SB_JOURNAL_BLOCK (s));

    /* fill by 1s the unused part of last bitmap */
    if (SB_BLOCK_COUNT (s) % (s->s_blocksize * 8))
	for (j = SB_BLOCK_COUNT (s) % (s->s_blocksize * 8); j < s->s_blocksize * 8; j ++)
	    set_bit (j, g_new_bitmap[SB_BMAP_NR (s) - 1]);

    /* allocate space for bitmap of uninsertable leaves */
    g_uninsertable_leaf_bitmap = getmem (sizeof (char *) * SB_BMAP_NR (s));
    for (i = 0; i < SB_BMAP_NR (s); i ++) {
	g_uninsertable_leaf_bitmap[i] = getmem (s->s_blocksize);
	memset (g_uninsertable_leaf_bitmap[i], 0xff, s->s_blocksize);
    }

    /* bitmap of formatted nodes */
    g_formatted = getmem (sizeof (char *) * SB_BMAP_NR (s));
    for (i = 0; i < SB_BMAP_NR (s); i ++) {
	g_formatted[i] = getmem (s->s_blocksize);
	memset (g_formatted[i], 0, s->s_blocksize);
    }
    /* bitmap of unformatted nodes */
    g_unformatted = getmem (sizeof (char *) * SB_BMAP_NR (s));
    for (i = 0; i < SB_BMAP_NR (s); i ++) {
	g_unformatted[i] = getmem (s->s_blocksize);
	memset (g_unformatted[i], 0, s->s_blocksize);
    }
}
#endif


/* write bitmaps and brelse them */
static void update_bitmap (struct super_block * s)
{
    int i;

    /* journal area could be used, reset it */
    for (i = 0; i < SB_JOURNAL_BLOCK (s) + rs_journal_size (s->s_rs) + 1; i ++)
	if (!is_block_used (i))
	    mark_block_used (i);

    for (i = 0; i < SB_BMAP_NR (s); i ++) {

	/* copy newly built bitmap to cautious bitmap */
	memcpy (SB_AP_BITMAP (s)[i]->b_data, g_new_bitmap[i], s->s_blocksize);

	mark_buffer_dirty (SB_AP_BITMAP (s)[i], 1);
	mark_buffer_uptodate (SB_AP_BITMAP (s)[i], 1);
	
	bwrite (SB_AP_BITMAP (s)[i]);

	freemem (g_new_bitmap[i]);
	/* g_disk_bitmap[i] points to corresponding cautious bitmap's b_data */
	freemem (g_uninsertable_leaf_bitmap[i]);
    }

    freemem (g_disk_bitmap);
    freemem (g_new_bitmap);
    freemem (g_uninsertable_leaf_bitmap);
}


static void release_bitmap (void)
{
    int i;

    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	brelse (SB_AP_BITMAP (&g_sb)[i]);
    }
}

static void release_super_block (void)
{
    bwrite (SB_BUFFER_WITH_SB (&g_sb));
    freemem (SB_AP_BITMAP (&g_sb));
    brelse (SB_BUFFER_WITH_SB (&g_sb));

    freemem (g_old_rs);
}


static inline int nothing_todo (struct super_block * s)
{
    if (opt_fsck)
	return 1;
    return 0;
}



#define WARNING \
"\nThis is an experimental version of reiserfsck, MAKE A BACKUP FIRST!\n\
Don't run this program unless something is broken.  You may want\n\
to backup first.  Some types of random FS damage can be recovered\n\
from by this program, which basically throws away the internal nodes\n\
of the tree and then reconstructs them.  This program is for use only\n\
by the desperate, and is of only beta quality.  Email\n\
reiserfs@devlinux.com with bug reports. \n"

/* 
   warning #2
   you seem to be running this automatically.  you are almost
   certainly doing it by mistake as a result of some script that
   doesn't know what it does.  doing nothing, rerun without -p if you
   really intend to do this.  */

void warn_what_will_be_done (struct what_to_do * job)
{
    char * answer = 0;
    size_t n = 0;
    __u32 dir_id, objectid;

    /* warn about fsck mode */
    switch (job->mode) {
    case FSCK_CHECK:
	fprintf (stderr, "Will read-only check consistency of the partition\n");
	break;

    case FSCK_MAKE_SB:
	fprintf (stderr, "Will check SB and rebuild if it is needed\n");
	break;

    case FSCK_REBUILD:
	fprintf (stderr, WARNING);
	break;
/*
    case FSCK_FAST_REBUILD:
	fprintf (stderr, WARNING);
	break;
*/
    case FSCK_SEARCH_KEY:
	fprintf (stderr, "Will look for the item with key\n");
	break;
    }
    if (!job->do_journal_replay)
	fprintf (stderr, "Skipping journal replay\n");


    if (job->mode == FSCK_REBUILD || !job->yes_confirmed) {
	fprintf (stderr, "Do you want to run this "
	    "program?[N/Yes] (note need to type Yes):");
	if (getline (&answer, &n, stdin) != 4 || strcmp ("Yes\n", answer))
	    exit (0);
    }
}


#if 0

void end_fsck (char * file_name)
{
    update_super_block ();
    update_bitmap (&g_sb);

    release_bitmap ();
    release_super_block ();

    if (opt_verbose == 1)
	output_information ();

    fprintf (stderr, "Syncing.."); fflush (stdout);
    
    flush_buffers ();
    sync ();
    
    fprintf (stderr, "done\n"); fflush (stdout);

    if (opt_verbose == 1)
	fprintf (stderr, "Checking mem..");
	
    free_overwritten_unfms ();
//    check_and_free_buffer_mem ();
    flush_buffers ();

    if (opt_verbose == 1)
	fprintf (stderr, "done\n");
    if (opt_fsck == 1)
	fprintf(stderr, "ReiserFS : done checking %s\n", file_name);
    else
	fprintf (stderr, "Ok\n");
    exit (0);
}


static void open_device (char * file_name, int flag)
{
    g_sb.s_dev = open (file_name, flag);
    if (g_sb.s_dev == -1)
	die ("reiserfsck: can not open '%s': %s", file_name, strerror (errno));
}  

static void reopen_read_only (char * file_name)
{
    close (g_sb.s_dev);
    open_device (file_name, O_RDONLY);
}

static void reopen_read_write (char * file_name)
{
    close (g_sb.s_dev);
    open_device (file_name, O_RDWR);
}

#endif


/* ubitmap.c: */extern int from_journal;



static void do_check (char * file_name, struct what_to_do * job)
{
    reiserfs_filsys_t fs;
    int error;

    fs = reiserfs_open (file_name, O_RDONLY, &error, 
			getmem (sizeof (struct fsck_union)));
    if (!fs) {
	fprintf (stderr, "\n\nreiserfsck: can not open reiserfs on \"%s\": %s\n\n",
		 file_name, error ? strerror (error) : "there is no one");
	return;
    }

    if (job->do_journal_replay)
	reiserfs_replay_journal (fs);
    
    fsck_union (fs)->fu_job = job;

    check_fs_tree (fs);

    semantic_check (fs);

    reiserfs_close (fs);
}


static void do_rebuild_sb (char * file_name)
{
    die ("do_rebuild_sb: not ready");
#if 0
    reiserfs_filsys * fs;

    fs = reiserfs_open (filename);
    rebuild_sb (fs);

    if (!fs)
	rebuild_sb (1);
    else
<<<<<<< main.c
	rebuild_sb (0);
=======
	rebuild_sb ();
>>>>>>> 1.13

    if (uread_super_block (&g_sb))
	rebuild_sb(1);
    else
	rebuild_sb(0);
<<<<<<< main.c
    reopen_read_write (filename);
=======

    reopen_read_write (filename);
>>>>>>> 1.13
    release_super_block ();
    flush_buffers ();
    exit (0);
#endif
}


/* fs has pointer to super block. Its contents is valid:
   s_block_count, s_bmap_nr, s_version, etc. Fields which will be
   changed by rebuilding: s_free_blocks, s_root_block, s_oid_cursize,
   s_tree_height, maybe s_hash_function_code. Reset them */
static void reset_super_block (reiserfs_filsys_t fs)
{
    /* reset few fields in super block */
    set_free_blocks (fs->s_rs, SB_BLOCK_COUNT (fs));
    if (fsck_union (fs)->fu_job->mode != FSCK_FAST_REBUILD){
        set_root_block (fs->s_rs, ~0);
        set_tree_height (fs->s_rs, ~0);
    }
    set_state (fs->s_rs, REISERFS_ERROR_FS);

    mark_buffer_dirty (SB_BUFFER_WITH_SB (fs), 0);
    /* make file system invalid unless fsck done () */
    bwrite (SB_BUFFER_WITH_SB (fs));

}


static void init_bitmaps (reiserfs_filsys_t fs)
{
  int i;

    /* make new bitmap and mark super block, bitmap and journal blocks
       used */
    new_bitmap (fs) = reiserfs_create_bitmap (SB_BLOCK_COUNT (fs));
    if (!new_bitmap (fs))
	die ("init_bitmaps: could not create new bitmap\n");

    for (i = 0; i <= SB_BUFFER_WITH_SB (fs)->b_blocknr; i ++)
	reiserfs_bitmap_set_bit (new_bitmap (fs), i);

    for (i = 0; i < SB_BMAP_NR (fs); i ++)
	reiserfs_bitmap_set_bit (new_bitmap (fs), i);

    for (i = 0; i < JOURNAL_BLOCK_COUNT + 1; i ++)
	reiserfs_bitmap_set_bit (new_bitmap (fs), i);

    /* bitmap of uninsertable leaves. fixme: should not be too many of
       these */
    uninsertable_leaves (fs) = reiserfs_create_bitmap (SB_BLOCK_COUNT (fs));
    reiserfs_bitmap_fill (uninsertable_leaves (fs));

    /* bitmap of unformatted nodes */
    unformatted_bitmap (fs) = reiserfs_create_bitmap (SB_BLOCK_COUNT (fs));

    /* bitmap of formatted nodes */
    formatted_bitmap (fs) = reiserfs_create_bitmap (SB_BLOCK_COUNT (fs));
    
    if (!uninsertable_leaves (fs) || !unformatted_bitmap (fs) ||
	!formatted_bitmap (fs))
	die ("init_bitmaps: could not create one of auxiliary bitmap");
	
}


static void do_rebuild (char * file_name, struct what_to_do * job)
{
    reiserfs_filsys_t fs;
    int error;

    fs = reiserfs_open (file_name, O_RDWR, &error,
			getmem (sizeof (struct fsck_union)));
    if (!fs) {
	fprintf (stderr, "\n\nreiserfsck: can not open reiserfs on \"%s\": %s\n\n",
		 file_name, error ? strerror (error) : "there is no one");
	return;
    }

    if (job->do_journal_replay)
	reiserfs_replay_journal (fs);

    fsck_union (fs)->fu_job = job;

    reset_super_block (fs);
    init_bitmaps (fs);


    /* 0 pass: look through all blocks and build leaf and unformatted maps */	
    pass0();

}


static void do_fast_rebuild (char * file_name)
{
    die ("do_fast_rebuild: not ready");
}


int main (int argc, char * argv [])
{
    char * file_name;
    int retval;
    struct what_to_do job = {FSCK_CHECK, 1/*replay journal*/,
			     0/*verbose*/, 0/*yes_confirmed*/, 0/*fix_fixable*/};


    print_banner ("reiserfsck");

    file_name = parse_options (argc, argv, &job);

    warn_what_will_be_done (&job); /* and ask confirmation Yes */

    switch (job.mode) {
    case FSCK_CHECK:
	do_check (file_name, &job);
	break;

    case FSCK_MAKE_SB:
	do_rebuild_sb (file_name);
	break;

    case FSCK_REBUILD:
	do_rebuild (file_name, &job);
	break;

    case FSCK_FAST_REBUILD:
	do_fast_rebuild (file_name);
	break;
    }
}

#if 0
	
/*    case FSCK_CHECK:*/
    case FSCK_REBUILD:
	
    }
    open_device (file_name, O_RDONLY);
    
    if (opt_fsck_mode == FSCK_SB)
    {
        if (uread_super_block (&g_sb))
          rebuild_sb(1);
        else
          rebuild_sb(0);
          reopen_read_write (file_name);
          release_super_block ();
          flush_buffers ();
          exit (0);
    }

    if (uread_super_block (&g_sb))
        die ("reiserfsck: no reiserfs found");

    g_th.t_super = &g_sb;
    //root_key.version = (SB_VERSION(&g_sb) == REISERFS_VERSION_2)?ITEM_VERSION_2:ITEM_VERSION_1;


    if (nothing_todo (&g_sb)) {
	/* this should work when fsck is called by fsck -a */
	printf ("%s: clean, %d/%d %ldk blocks\n", file_name,
		SB_BLOCK_COUNT (&g_sb) - SB_FREE_BLOCKS(&g_sb), SB_BLOCK_COUNT (&g_sb), g_sb.s_blocksize / 1024);
	brelse (SB_BUFFER_WITH_SB (&g_sb));
	return 0;
    }


    if (opt_fsck_mode == FSCK_DEFAULT || opt_fsck_mode == FSCK_CORRECT_BITMAP) {
	check_fs_tree (&g_sb);
	release_bitmap ();
	release_super_block ();
//	check_and_free_buffer_mem ();
	reopen_read_write (file_name);
	flush_buffers ();
	exit (0);
    }

    if (opt_stop_point == STOP_AFTER_REPLAY) {
	release_super_block ();
//	check_and_free_buffer_mem ();
	flush_buffers ();
	//reiserfs_sync_buffers(0,0);
	exit (0);	
    }
	
    if (opt_fsck_mode == FSCK_SEMANTIC)
    {
        semantic_check();
//	release_bitmap ();
	release_super_block ();
	flush_buffers ();
	exit (0);
    }

    if (opt_fsck_mode == FSCK_FAST_REBUILD)
    {
	__u32 root_block = SB_ROOT_BLOCK(&g_sb);
	reopen_read_write (file_name);
	fprintf (stderr, "Replaying log..");
	replay_all (&g_sb);
	fprintf (stderr, "done\n");
	if (opt_fsck == 1)
	    fprintf (stderr, "ReiserFS : checking %s\n",file_name);
	else
	    fprintf (stderr, "Rebuilding..\n");

	
	reset_super_block (&g_sb);
	SB_DISK_SUPER_BLOCK(&g_sb)->s_root_block = cpu_to_le32 (root_block);
	init_bitmaps (&g_sb);

	/* 1,2. building of the tree */
	recover_internal_tree(&g_sb);

	/* 3. semantic pass */
	semantic_pass ();

	/* if --lost+found is set - link unaccessed directories to lost+found directory */
	pass4 (&g_sb);
	compare_id_maps(&g_sb);
	free_id_map();

	/* 4. look for unaccessed items in the leaves */
	check_unaccessed_items ();
	
	if (from_journal) {
	    printf ("%d blocks from journal area [%d %d] has been used to perform repairing. "
	    "Will release them. This may take a while\nScanning tree..",
		from_journal, SB_JOURNAL_BLOCK (&g_sb),
		SB_JOURNAL_BLOCK (&g_sb) + rs_journal_size (g_sb.s_rs));
	    pass_through_tree (&g_sb, release_journal_blocks, 0);
	}

	end_fsck(file_name);
    }

    if (opt_fsck_mode == FSCK_REBUILD) {
	reopen_read_write (file_name);

	fprintf (stderr, "Replaying log..");
	replay_all (&g_sb);
	fprintf (stderr, "done\n");

	if (opt_fsck == 1)
	    fprintf (stderr, "ReiserFS : checking %s\n",file_name);
	else
	    fprintf (stderr, "Rebuilding..\n");

	reset_super_block (&g_sb);
	init_bitmaps (&g_sb);

	/* 0 pass: look through all blocks and build leaf and unformatted maps */	
	pass0();
	
	/* 1,2. building of the tree */
	build_the_tree ();

	/* 3. semantic pass */
	semantic_pass ();
	
	/* if --lost+found is set - link unaccessed directories to lost+found directory */
	pass4 (&g_sb);
	compare_id_maps(&g_sb);
	free_id_map();

	/* 4. look for unaccessed items in the leaves */
	check_unaccessed_items ();
	
	if (from_journal) {
	    printf ("%d blocks from journal area [%d %d] has been used to perform repairing. "
	    "Will release them. This may take a while\nScanning tree..",
		from_journal, SB_JOURNAL_BLOCK (&g_sb),
		SB_JOURNAL_BLOCK (&g_sb) + rs_journal_size (g_sb.s_rs));
	    pass_through_tree (&g_sb, release_journal_blocks, 0);
	}

	end_fsck (file_name);
    }

    if (opt_fsck_mode == FSCK_FIND_ITEM) {
	init_bitmaps (&g_sb);
	build_the_tree ();
	release_bitmap ();
	release_super_block ();
//	check_and_free_buffer_mem ();
	flush_buffers ();
	exit (0);
    }


    return 0;
}
#endif
