/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "fsck.h"
#include <stdlib.h>

/* This file contains functions for 4 passes - pass0, pass1, pass1a, and
   pass2.
  
   Pass0: read all blocks marked used, find blocks looking like leaves, mark
          them as formatted (found_leaves bitmap) and all blocks pointed to by
          indirect items as unformatted (unformatted_bitmap). Indirect item
          slots pointing to block marked free in on-disk bitmap are zeroed.

	  After scanning - 
	  mark blocks pointed by both bitmaps as uninsertable
	  (not_for_insertion bitmap) 
	  
	  mark internals free (zero coresponding bits in disk_bitmap) 

	  reset unformatted bitmap to zeros - this is needed to find duplicate
	  unformatted pointers.  If we didn't zero unfm bitmap we won't be
	  able catch all duplicate blocks: insertion of block can be postponed
	  and we must be sure that there is no another indirect node pointing
	  to it.it will be used )

   Pass1: read blocks marked in formatted bitmap and not marked in not_for_insertion bitmap

   looking through formatted blocks only and trying insert them into tree.
If block data are not correct, save all good items and unmarked block as formatted
If block is uninsertable (this leaf is pointed to by indirect item), save all good
items and unmarked block as formatted and unmarked block as uninsertable
If block data are correct, but this block cannot be inserted into tree, postpone its
insertion - mark block as uninsertable
For all these cases we handle indirect items in pointers_in_pass1 which duplicates
content of whole pointed block to the new one if this is a pointer to leaf or a
duplicate pointer.
Get new buffer from blocks which are !was_used and !is_used in g_new_bitmap.
For have saved and postoned blocks mark all pointed blocks as was_used.
Pass1a:
    trying to put_saved_items_into_tree
Pass 2:
    trying to put items from uninsertable blocks into tree

All found bad blocks are marked as used, not formatted and not unformatted */


/* step == 	0 - find stat_data or any item ;
		1 - find item ;
		2 - already found */

/* new algorithm: in pass 0 : 
   read used blocks.
   mark: 
   1. perfectly correct leaves (maybe fixed),
   2. unformatted nodes pointed by indirect items,
   3. creating list of unformatted nodes pointed by more then one leaf
   4. leaves pointed to by indirect items (these are correct, however the question is: should we put those items separately in the tree? No.)

   at the end of pass0 - find which leaves are not pointed by indirect items

   pass1: create tree of those good leaves zeroing identical pointers.

   
 */

int step = 0;

typedef void (action_on_pointers_t)(struct buffer_head *, struct item_head *, int);

inline void pointers_in_pass0(struct buffer_head * bh, struct item_head * ih, int flag)
/* function for pass0: we do not correct items in pass0, because we have
   to read and check them again in pass1. */
{
    __u32 * ind;
    int i;

    if (is_indirect_ih (ih))
        for (i = 0, ind = (__u32 *)B_I_PITEM (bh,ih); i < I_UNFM_NUM (ih); i++, ind++){

            if (*ind == 0)
                continue;
	
            if (*ind >= SB_BLOCK_COUNT (&g_sb)) /* invalid pointer */
                continue;

            if (!was_block_used (*ind)) {       /* block is marked free in disk bitmap */
                if (opt_verbose)
		    printf("\npointers_in_pass0: pointer (%lu->%lu) to unused part of partition - delete ", bh->b_blocknr, *ind);
                *ind = 0;                       /* this block can be allocated later, so we should correct it here */
                mark_buffer_dirty (bh, 1);
//     	        mark_buffer_uptodate (bh, 1);
                continue;
            }

            if (!is_block_unformatted(*ind))
                mark_block_unformatted(*ind);
        }
}

inline void pointers_in_pass1(struct buffer_head * bh, struct item_head * ih, int flag)
{
    int i;
    __u32 * ind;

    if (is_indirect_ih (ih))
        /* check each pointer to unformatted node, if it is in the tree already, put 0 here */
	for (i = 0, ind = (__u32 *)B_I_PITEM(bh,ih); i < I_UNFM_NUM (ih); i ++, ind++) {

            if (*ind == 0)
                continue;
	
            if (!was_block_used (*ind))       /* block is marked free on disk bitmap, this should be corrected in pass0 */
                die("pointers_in_pass1: pointer to (%d) must be used in cautious bitmap. block %d", *ind, bh->b_blocknr);

            if (*ind >= SB_BLOCK_COUNT (&g_sb)) { /* invalid data block */
                *ind = 0;
                if (flag != SAVE)
                    mark_buffer_dirty (bh, 1);
                continue;
            }

            /* this means *ind points to internal or to leaf in the tree. It cannot be possible because */
            /* 1. new block we shoose from !was_block_used blocks. After pass0 all pointed blocks
            /*    became was_block_used
            /* 2. all blocks pointed to as unfm are being saving in handle_leaf_in_pass1 */
            if (is_block_used(*ind) && !is_block_unformatted(*ind))
		die("pointers_in_pass1: cannot be used and !unformatted %lu", *ind);

            if (is_block_unformatted(*ind) || is_block_formatted (*ind))
            {/* is_block_unformatted - we met such a pointer already (it can be saved or postponed)
		is_block_formatted - saved or postponed leaf
		duplicate content of pointed block */
                struct buffer_head * bh_dest, * bh_src;
		
                if (is_block_formatted (*ind) && is_block_unformatted(*ind))
		    die("pointers_in_pass1: cannot be formatted and unformatted %lu", *ind);
		
		if (opt_verbose && is_block_unformatted(*ind))
		    printf("\npointers_in_pass1: pointer (%lu->%lu) to unformatted block was met already - copy ", bh->b_blocknr, *ind);
		
		if (opt_verbose && is_block_formatted(*ind))
		    printf("\npointers_in_pass1: pointer (%lu->%lu) to formatted block was met already - copy ", bh->b_blocknr, *ind);

                bh_src = bread (g_sb.s_dev, *ind, g_sb.s_blocksize);
                if (bh_src == 0)
                { /* all bad blocks should be used, not formatted and not unformatted */
                    if (is_block_unformatted(*ind))
                        unmark_block_unformatted(*ind);
                    if (is_block_formatted(*ind))
                        unmark_block_formatted(*ind);
                    if (!is_block_used(*ind))
                        mark_block_used(*ind);
                    *ind = 0;
                    if (flag != SAVE)
                        mark_buffer_dirty (bh, 1);
                    continue;
                }

                bh_dest = reiserfsck_get_new_buffer (bh->b_blocknr); /*bh_dest marked as used*/
                if (bh_dest) {
                    /* this block must be free in new bitmap if it is not INSERT */
                    mark_block_free(bh_dest->b_blocknr);
                    /* whan we allocate new block we choose it from !was_used blocks so we should mark it as was_used */
                    /* this block can be marked as was_used already if it is journal blocks */
                    mark_block_was_used(bh_dest->b_blocknr);

                    memcpy (bh_dest->b_data, bh_src->b_data, bh_src->b_size);
                    mark_buffer_dirty (bh_dest, 1);
                    mark_buffer_uptodate (bh_dest, 1);

                    *ind = bh_dest->b_blocknr;
                    brelse (bh_dest);
                    if (flag != SAVE) {
                        mark_buffer_dirty (bh, 1);
//                      mark_buffer_uptodate (bh, 1);
                    }
                }else   /* no free disk space */
                        /* FIXME: probably we should zero the 2 pointers on the same block */
                        /* flag must be not INSERT not to mark block as used */
                    flag = SAVE;

                brelse (bh_src);
            }
            /* ok, mark that block is in tree and that it is unformatted node */
            if (flag == INSERT) {
                mark_block_used (*ind);
            }	
	    /* mark all pointed blocks as unfm even if this leaf is postponed */
	    /* because there can be another pointer to this block */	
            mark_block_unformatted (*ind);

	}
}

/* leaf data can be bad, check every item before handling it */
void for_all_good_items_in_node (action_on_item_t action, struct si ** si, action_on_pointers_t action2,
                                struct buffer_head * bh, int flag)
{
    int i;
    struct item_head * ih;

    for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++)
	if (!is_bad_item (bh, B_I_PITEM (bh,ih), g_sb.s_blocksize, g_sb.s_dev)) {
	    action2(bh, ih, flag);
	    if (action)
		action (si, ih, B_I_PITEM(bh,ih));
	}
    if (action) /*SAVE*/
	unmark_block_formatted(bh->b_blocknr);
}

/* it is for good leaves only */
void for_all_items_in_node (action_on_item_t action, struct si ** si, struct buffer_head * bh, int flag)
{
  int i;
  struct item_head * ih;

  for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++) {
      if (flag != INSERT_SEPARATELY) /* INSERT or POSTPONE or SAVE */
          pointers_in_pass1(bh, ih, flag);
      if (flag == SAVE || flag == INSERT_SEPARATELY)
          action (si, ih, B_I_PITEM(bh,ih));
  }
  if (flag == SAVE || flag == INSERT_SEPARATELY)
    unmark_block_formatted(bh->b_blocknr);
}

#if 0
void was_in_handle_indirect_items(struct buffer_head * bh, struct item_head * ih, int flag)
{
    int i;
    __u32 * ind = (__u32 *)B_I_PITEM(bh,ih);

    if (is_indirect_ih(ih)) 	
        /* check each pointer to unformatted node, if it is in the tree already, put 0 here */
	for (i = 0, ind = (__u32 *)B_I_PITEM(bh,ih); i < ih_item_len(ih) / UNFM_P_SIZE; i ++, ind++)
        {
            if (*ind == 0)
                continue;
            if (!was_block_used (*ind))   /* block is marked free on disk bitmap, this should be corrected in pass0 */
                die("handle_indirect_items: pointer must be used in cautious bitmap");
		
            if (*ind >= SB_BLOCK_COUNT (&g_sb)) {/* invalid data block */
                *ind = 0;
                mark_buffer_dirty (bh, 1);
                continue;
            }

            if (is_block_used(*ind))
            {/* it pointed to formatted block or we met such a pointer already -
                duplicate content of unfm block */
                struct buffer_head * bh_src, * bh_dest;

                if (!(bh_src = bread (g_sb.s_dev, *ind, g_sb.s_blocksize))) {
                    if (is_block_unformatted(*ind))
                        unmark_block_unformatted(*ind);
                    if (is_block_formatted(*ind))
                        unmark_block_formatted(*ind);
                    if (!is_block_used(*ind))
                        mark_block_used(*ind);
                    *ind = 0;
                    mark_buffer_dirty (bh, 1);
                    continue;
                }

                bh_dest = reiserfsck_get_new_buffer (bh->b_blocknr); /*bh_dest marked as used*/
//                mark_block_free(bh_dest->b_blocknr);
                mark_block_was_used(bh_dest->b_blocknr);
                memcpy (bh_dest->b_data, bh_src->b_data, bh_src->b_size);

                /* this block must be free in new bitmap */
                /* whan we allocate new block we choose it from !was_used blocks so we should mark it as was_used */
                mark_buffer_dirty (bh_dest, 1);
                mark_buffer_uptodate (bh_dest, 1);

                *ind = bh_dest->b_blocknr;
	
                brelse (bh_src);
                brelse (bh_dest);

                mark_buffer_dirty (bh, 1);
                mark_buffer_uptodate (bh, 1);
            }
            /* ok, mark that block is in tree and that it is unformatted node */
            if (flag == INSERT) {
                mark_block_used (*ind);
                mark_block_unformatted (*ind);
            }
        }
}

/* analyse contents of indirect items. If it points to used blocks or
   to uninsertable node, which has to be inserted by items - we copy
   content of those blocks and mark pointed blocks as used */
static void handle_indirect_items (struct buffer_head * bh, int flag)
{
    int i;
    struct item_head * ih;

    for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++)
        was_in_handle_indirect_item (bh, ih, flag);
}

void for_items (action_on_item_t action, struct si ** si, action_on_pointers_t action2,
                struct buffer_head * bh, int flag)
{
  int i;
  struct item_head * ih;

  for (i = 0, ih = B_N_PITEM_HEAD (bh, 0); i < B_NR_ITEMS (bh); i ++, ih ++)
      if (!is_bad_item (bh, B_I_PITEM (bh,ih), g_sb.s_blocksize, g_sb.s_dev))
          action2(bh, ih, flag);
//          pointers_in_pass0(bh, ih);
}
#endif

static void mark_bad_pointed_blocks_as_uninsertable()
{/* mark formatted & unformatted blocks as uninsertable */
    int i,j;

    for (i = 0; i < SB_BMAP_NR (&g_sb); i++)
        for (j = 0; j < g_sb.s_blocksize; j++)
            g_uninsertable_leaf_bitmap[i][j] = ~(g_unformatted[i][j] & g_formatted[i][j]);

    /* we need only blocks which are either formatted or unformatted */
    /* so, mark them as was_used - this can free some blocks for SCAN_WHOLE_PARTITION */
    set_free_blocks(g_old_rs, SB_BLOCK_COUNT (&g_sb));
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++)
        for (j = 0; j < g_sb.s_blocksize; j++) {
            g_disk_bitmap[i][j] = (g_unformatted[i][j] | g_formatted[i][j]);
            set_free_blocks(g_old_rs, rs_free_blocks(g_old_rs) + 8 - test_byte(&g_disk_bitmap[i][j]));
        }

    /* mark metadata and unused spase(before SB and the end of the last bitmap) as used again */
    /* skipped and super block */
    for (i = 0; i <= SB_BUFFER_WITH_SB (&g_sb)->b_blocknr; i ++)
	mark_block_was_used (i);

    /* bitmaps */
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++)
	mark_block_was_used (SB_AP_BITMAP (&g_sb)[i]->b_blocknr);

    /* journal */
    for (i = 0; i < rs_journal_size (g_old_rs) + 1; i ++)
	mark_block_was_used (i + SB_JOURNAL_BLOCK (&g_sb));

    /* unused space of last bitmap is filled by 1s */
    for (i = SB_BMAP_NR (&g_sb) * g_sb.s_blocksize * 8; --i >= SB_BLOCK_COUNT (&g_sb); ) {
	mark_block_was_used (i);
	set_free_blocks (g_old_rs, rs_free_blocks (g_old_rs) + 1);
    }

    /* to find duplicate unfm pointers we should zero unformatted bitmap  */
    /* insertion of block can be postponed and we must be sure that there */
    /* is no another indirect node pointing to it.			  */
    for (i = 0; i < SB_BMAP_NR (&g_sb); i++)
	memset (g_unformatted[i], 0, g_sb.s_blocksize);
}

/* allocates buffer head and copy buffer content */
static struct buffer_head * make_buffer (int dev, int blocknr, int size, char * data)
{
  struct buffer_head * bh;

  bh = getblk (dev, blocknr, size);
  if (buffer_uptodate (bh))
    return bh;
//    die ("make_buffer: uptodate buffer found");
  memcpy (bh->b_data, data, size);
  set_bit (BH_Uptodate, (char *)&bh->b_state);
  return bh;
}

/* 1 if these keys cannot be of one file */
/* to_find can contain -1 for everything in any field */
int find_not_of_one_file(struct key * to_find, struct key * key)
{
    if ((to_find->k_objectid != -1) &&
        (to_find->k_objectid != key->k_objectid))
        return 1;
    if ((to_find->k_dir_id != -1) &&
        (to_find->k_dir_id != key->k_dir_id))
        return 1;
    return 0;
}

static void find_a_key (struct key * key, __u64 offset, struct buffer_head * bh)
{
  int i;
  struct item_head * ih;
  struct reiserfs_de_head * deh;

  if (step == 2)
    return;

  for (i = 0; i < B_NR_ITEMS (bh); i ++)
  {
    ih = B_N_PITEM_HEAD (bh, i);
    if (find_not_of_one_file (key, &(ih->ih_key)))
      continue;

    if ( offset == MAX_OFFSET_V2 )
    {
        /* look for all items of this file */
        reiserfs_warning ("\nblock %d contains item of this file %k (item %d)\n", bh->b_blocknr, &(ih->ih_key), i);
            continue;
    }

    if (get_offset(&ih->ih_key) == offset)
    {
        reiserfs_warning ("\nblock %d contains key %k (item %d)\n", bh->b_blocknr, &(ih->ih_key), i);
        if ((key->k_objectid != -1) && (key->k_dir_id != -1))
        {
            /* key is specified precisely */
            step = 2;
            return;
        }
        continue;
    }

    if (is_direntry_ih (ih))
    {
      int j;

      for (j = 0, deh = B_I_DEH (bh, ih); j < ih_entry_count(ih); j ++, deh ++)
	if (deh->deh_offset == offset)
	{
	  reiserfs_warning ("\nblock %d contains key [%d %d %d 3] (item_pos %d, pos_in_item %d)\n",
	                                bh->b_blocknr, ih->ih_key.k_dir_id, ih->ih_key.k_objectid, offset, i, j);
          if ((key->k_objectid != -1) && (key->k_dir_id != -1))
          {
            /* key is specified precisely */
            step = 2;
  	    return;
  	  }
  	  continue;
	}
    }
  }
}

/*

static void find_a_key (struct key * key, __u64 offset, struct buffer_head * bh)
{
  int i;
  struct item_head * ih;
  struct reiserfs_de_head * deh;

  if (step == 2)
        return;

  for (i = 0; i < B_NR_ITEMS (bh); i ++) {
    ih = B_N_PITEM_HEAD (bh, i);
    if (not_of_one_file (key, &(ih->ih_key)))
      continue;

    if (!step)
    {
        if ( ((ih_version(ih) == ITEM_VERSION_2) && (offset == MAX_OFFSET_V2)) ||
             ((ih_version(ih) == ITEM_VERSION_1) && (offset == MAX_OFFSET_V1)) ) {
          // look for all items of this file
          reiserfs_warning ("\nblock %d contains item of this file %k (item %d)\n", bh->b_blocknr, key, i);
          return;
        }

        if (comp_le_keys (key, &(ih->ih_key)))
          continue;

        set_le_key_k_offset(ih_version(ih), key, offset);
        set_le_key_k_offset(ih_version(ih), key, TYPE_ANY);
        step = 1;
    }

    // key is specified precisely
    if (!comp_keys_3 (key, &(ih->ih_key), ih_version(ih))) {
      reiserfs_warning ("\nblock %d contains key %k (item %d)\n", bh->b_blocknr, key, i);
      step = 2;
      return;
    }
    if (is_direntry_le_key(ITEM_VERSION_1, key) && is_direntry_le_ih (ih)) {
      int j;
      
      deh = B_I_DEH (bh, ih);
      for (j = 0; j < I_ENTRY_COUNT (ih); j ++, deh ++)
	if (deh->deh_offset == le_key_k_offset(ITEM_VERSION_1, key)) {
	  reiserfs_warning ("\nblock %d contains key %k (item_pos %d, pos_in_item %d)\n", bh->b_blocknr, key, i, j);
          step = 2;
	  return;
	}
    }
  }
}
*/


int g_unaccessed_items = 0;

/* these are 4 bits only */
#define ITEM_REACHABLE 0
#define ITEM_UNREACHABLE 0xf

int is_item_reachable (struct item_head * ih)
{
    return (ih->ih_format.fsck_need == ITEM_REACHABLE) ? 1 : 0;
}


void mark_item_unreachable (struct item_head * ih)
{
    g_unaccessed_items ++;
    ih->ih_format.fsck_need = ITEM_UNREACHABLE;
//&&&&&&&&&&&&&&&&&&&&&&&&&&
    if (is_indirect_ih (ih))
      set_free_space (ih, 0);
//&&&&&&&&&&&&&&&&&&&&&&&&&&
    //  we do not need mark_buffer_dirty here because there is no way
    //  for this item to get into not dirtied node
}


void mark_item_reachable (struct item_head * ih, struct buffer_head * bh)
{
    g_unaccessed_items --;
    ih->ih_format.fsck_need = ITEM_REACHABLE;
    mark_buffer_dirty (bh, 0);
}


/* used when leaf is inserted into tree by pointer
   1. set sd_nlinks to 0 in all stat data items
   2. mark all items as unaccessed
   */
static void reset_nlinks (struct buffer_head * bh)
{ 
    int i;
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < B_NR_ITEMS (bh); i ++, ih ++) {
	mark_item_unreachable (ih);
	if (is_stat_data_ih (ih)) {
	    add_event (STAT_DATA_ITEMS);
            if (ih_key_format(ih) == KEY_FORMAT_1)
                ((struct stat_data_v1 *)B_I_PITEM(bh,ih))->sd_nlink = 0;
            else
                ((struct stat_data *)B_I_PITEM(bh,ih))->sd_nlink = 0;
	}
    }
    mark_buffer_dirty (bh, 0);
}


/* ubitmap.c */
extern int from_journal;

static void insert_pointer (struct buffer_head * bh, struct path * path)
{
  struct item_head * ih;
  char * body;
  int zeros_number;
  int retval;
  struct tree_balance tb;

  init_tb_struct (&tb, &g_sb, path, 0x7fff);
//  tb.preserve_mode = NOTHING_SPECIAL;

  /* fix_nodes & do_balance must work for internal nodes only */
  ih = 0;
//  retval = fix_nodes (0, M_INTERNAL, &tb, PATH_LAST_POSITION (path), ih);

  retval = fix_nodes (tb.transaction_handle, M_INTERNAL, &tb, ih);

  if (retval != CARRY_ON)
    die ("insert_pointer: no free space on device (retval == %d, used blocks from journal %d",
	 retval, from_journal);

  /* child_pos: we insert after position child_pos: this feature of the insert_child */
  /* there is special case: we insert pointer after
     (-1)-st key (before 0-th key) in the parent */
  if (PATH_LAST_POSITION (path) == 0 && path->pos_in_item == 0)
    PATH_H_B_ITEM_ORDER (path, 0) = -1;
  else {
    if (PATH_H_PPARENT (path, 0) == 0)
      PATH_H_B_ITEM_ORDER (path, 0) = 0;
/*    PATH_H_B_ITEM_ORDER (path, 0) = PATH_H_PPARENT (path, 0) ? PATH_H_B_ITEM_ORDER (path, 0) : 0;*/
  }

  ih = 0;
  body = (char *)bh;
  //memmode = 0;
  zeros_number = 0;

//  do_balance (0, &tb, path->pos_in_item, ih, body, M_INTERNAL, memmode, zeros_number);
  do_balance (0/*th*/, &tb, ih, body, M_INTERNAL, zeros_number);

  /* mark as used block itself and pointers to unformatted nodes */
  mark_block_used (bh->b_blocknr);

  /* this is for check only */
//  mark_block_formatted (bh->b_blocknr);
  reset_nlinks (bh);

//  handle_indirect_items (bh, INSERT);
  for_all_items_in_node (0, 0, bh, INSERT);

  /* statistic */
  add_event (GOOD_LEAVES);
}


/* return 1 if left and right can be joined. 0 otherwise */
// FIXME: note, that tree is not completed yet, that is left and right
// not necessarily will be neighbors at the final tree, that is they
// do not need to satisfy the balance condition
int balance_condition_fails (struct buffer_head * left, struct buffer_head * right)
{
  if (B_FREE_SPACE (left) >= B_CHILD_SIZE (right) -
      (are_items_mergeable (B_N_PITEM_HEAD (left, B_NR_ITEMS (left) - 1), B_N_PITEM_HEAD (right, 0), left->b_size) ? IH_SIZE : 0))
    return 1;
  return 0;
}


/* return 1 if new can be joined with last node on the path or with
   its right neighbor, 0 otherwise */
int balance_condition_2_fails (struct buffer_head * new, struct path * path)
{
  struct buffer_head * bh;
  struct key * right_dkey;
  int pos, used_space;

  bh = PATH_PLAST_BUFFER (path);


  if (balance_condition_fails (bh, new))
    /* new node can be joined with last buffer on the path */
    return 1;

  /* new node can not be joined with its left neighbor */

  right_dkey = uget_rkey (path);
  if (right_dkey == 0)
    /* there is no right neighbor */
    return 0;
  
  pos = PATH_H_POSITION (path, 1);
  if (pos == B_NR_ITEMS (bh = PATH_H_PBUFFER (path, 1))) {
    /* we have to read parent of right neighbor. For simplicity we
       call search_by_key, which will read right neighbor as well */
//    init_path (&path_to_right_neighbor);
    INITIALIZE_PATH(path_to_right_neighbor);

    if (usearch_by_key (&g_sb, right_dkey, &path_to_right_neighbor) != ITEM_FOUND)
      die ("get_right_neighbor_free_space: invalid right delimiting key");
    used_space =  B_CHILD_SIZE (PATH_PLAST_BUFFER (&path_to_right_neighbor));
    pathrelse (&path_to_right_neighbor);
  }
  else
    used_space = B_N_CHILD (bh, pos + 1)->dc_size;
  
  if (B_FREE_SPACE (new) >= used_space -
       (are_items_mergeable (B_N_PITEM_HEAD (new, B_NR_ITEMS (new) - 1), (struct item_head *)right_dkey, new->b_size) ? IH_SIZE : 0))
    return 1;

  return 0;
}

static void get_max_buffer_key (struct buffer_head * bh, struct key * key)
{
    struct item_head * ih;

    ih = B_N_PITEM_HEAD (bh, B_NR_ITEMS (bh) - 1);
    copy_key (key, &(ih->ih_key));

    if (is_direntry_key (key)) {
	/* copy 3-rd and 4-th key components of the last entry */
	//set_le_key_k_offset (ih_version(ih), key, B_I_DEH (bh, ih)[I_ENTRY_COUNT (ih) - 1].deh_offset);
	//set_le_key_k_type (ih_version(ih), key, TYPE_DIRENTRY);
	set_offset (KEY_FORMAT_1, key, 
			     le32_to_cpu (B_I_DEH (bh, ih)[ih_entry_count (ih) - 1].deh_offset));

    } else if (!is_stat_data_key (key))
	/* get key of the last byte, which is contained in the item */
	set_offset (key_format (key), key, get_offset (key) + get_bytes_number (ih, bh->b_size) - 1);
    //set_le_key_k_offset(ih_version(ih), key,
    //		    le_key_k_offset(ih_version(ih), key) + get_bytes_number (bh, ih, 0, CHECK_FREE_BYTES) - 1 );
}


/* inserts pointer to leaf into tree if possible. If not, marks node as uninsrtable */
static void try_to_insert_pointer_to_leaf (struct buffer_head * new_bh)
{
  struct path path;
  struct buffer_head * bh;			/* last path buffer */
  struct key * first_bh_key, last_bh_key;	/* first and last keys of new buffer */
  struct key last_path_buffer_last_key, * right_dkey;
  int ret_value;


  path.path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
  if (is_block_used (new_bh->b_blocknr))
    /* block could get into tree already if its number was used by
       some indirect item */
    goto cannot_insert;

  first_bh_key = B_N_PKEY (new_bh, 0);

  /* try to find place in the tree for the first key of the coming node */
/*
  if (KEY_IS_DIRECTORY_KEY (first_bh_key))
    ret_value = search_by_entry_key (&g_sb, first_bh_key, &path);
  else if (KEY_IS_STAT_DATA_KEY (first_bh_key)) {
    ret_value = reiserfsck_search_by_key (&g_sb, first_bh_key, &path, comp_keys);
    path.pos_in_item = 0;
  } else
    ret_value = search_by_position (&g_sb, first_bh_key, &path);
  if (ret_value == KEY_FOUND || ret_value == ENTRY_FOUND || ret_value == BYTE_FOUND)
    goto cannot_insert;
*/
  ret_value = usearch_by_key (&g_sb, first_bh_key, &path);
  if (ret_value == ITEM_FOUND)
    goto cannot_insert;

  /* get max key in the new node */
  get_max_buffer_key (new_bh, &last_bh_key);

  bh = PATH_PLAST_BUFFER (&path);
  if (comp_keys (B_N_PKEY (bh, 0), &last_bh_key) == 1/* first is greater*/) {
    /* new buffer falls before the leftmost leaf */
    if (balance_condition_fails (new_bh, bh))
      goto cannot_insert;

    if (uget_lkey (&path) != 0 || PATH_LAST_POSITION (&path) != 0)
      die ("try_to_insert_pointer_to_leaf: bad search result");

    path.pos_in_item = 0;
    goto insert;
  }

  /* get max key of buffer, that is in tree */
  get_max_buffer_key (bh, &last_path_buffer_last_key);
  if (comp_keys (&last_path_buffer_last_key, first_bh_key) != -1/* second is greater */)
    /* first key of new buffer falls in the middle of node that is in tree */
    goto cannot_insert;

  right_dkey = uget_rkey (&path);
  if (right_dkey && comp_keys (right_dkey, &last_bh_key) != 1 /* first is greater */)
    goto cannot_insert;

  if (balance_condition_2_fails (new_bh, &path))
    goto cannot_insert;

insert:
  insert_pointer (new_bh, &path);
  goto out;

cannot_insert:
  /* statistic */
  add_event (UNINSERTABLE_LEAVES);
//  handle_indirect_items (new_bh, POSTPONE);
  for_all_items_in_node (0, 0, new_bh, POSTPONE);
  if (opt_verbose)
      reiserfs_warning ("\ntry_to_insert_pointer_to_leaf: leaf is postponed: %lu\n", new_bh->b_blocknr);
  mark_block_uninsertable (new_bh->b_blocknr);

out:
  pathrelse (&path);
  brelse (new_bh);
  return;
}


static int tree_is_empty (void)
{
  return (SB_ROOT_BLOCK (&g_sb) == ~0) ? 1 : 0;
}


static void make_single_leaf_tree (struct buffer_head * bh)
{
    struct reiserfs_super_block * rs;
     
    rs = (struct reiserfs_super_block *)(SB_BUFFER_WITH_SB (&g_sb)->b_data);

    /* tree is empty, make tree root */
    set_root_block (rs, bh->b_blocknr);
    set_tree_height (rs, 2);

    mark_block_used (bh->b_blocknr);
    mark_buffer_dirty (SB_BUFFER_WITH_SB (&g_sb), 1);

    /* this is for check only */
//    mark_block_formatted (bh->b_blocknr);
    
    /* set stat data nlinks fields to 0, mark all items as unaccessed,
       analyse contents of indirect items */
    reset_nlinks (bh);
//    handle_indirect_items (bh, INSERT);
    for_all_items_in_node (0, 0, bh, INSERT);

    /* statistic */
    add_event (GOOD_LEAVES);
    
    brelse (bh);
}

__u32 cur_pos; 			/* current position */

#define NR_TO_READ      8 	/* read only by 8 blocks now */

/* read one/some blocks functions can return these exceptions */
#define OUT_OF_BITMAP   1
#define NOT_USED        2
#define LAST_BITS       3
#define BAD_BLOCK       4


/* were [pos, pos + NR_TO_READ] blocks used in on-disk bitmap */
int were_blocks_used(__u32 pos)
{
    int i;

    for (i = 0; i<NR_TO_READ; i++)
        if (was_block_used(pos + i))
            return 1;

    return 0;
}

/* any of [pos, pos + NR_TO_READ] blocks is unused and formatted */
int are_new_blocks_formatted(__u32 pos)
{
    int i;

    for (i = 0; i < NR_TO_READ; i++)
        if (is_block_formatted(pos + i) && !is_block_used(pos + i))
            return 1;

    return 0;
}

typedef int (check_block_rules_t) (__u32);

/* all rules return 1 when we shouldn't read/check block(s) */
int one_block_rule_for_pass0(__u32 pos)
{/* rule for read_one_block function in pass0,
  * block wasn't used and we are scanning used part only */

    if (not_data_block (&g_sb, pos))
        return 1; /* skip not data area of the filesystem (journal, bitmaps, reserved space) */

    return (!was_block_used(pos));
}

int one_block_rule_for_pass1(__u32 pos)
{/* rule for read_one_block function in pass1,
  * block isn't formatted */

    if (not_data_block (&g_sb, pos))
        return 1; /* skip not data area of the filesystem (journal, bitmaps, reserved space) */

    return (!is_block_formatted (pos) || is_block_used(pos));
}


int some_blocks_rule_for_pass0(__u32 pos)
{/* rule for read_some_blocks function in pass0,
  * blocks aren't used and we are scaning used part only */

    if (pos + NR_TO_READ < SB_JOURNAL_BLOCK (&g_sb) + JOURNAL_BLOCK_COUNT + 1)
	return 1; /* skip not data area of the filesystem (journal, reserved space) */

    return (!were_blocks_used(pos));
}

int some_blocks_rule_for_pass1(__u32 pos)
{/* rule for read_some_blocks function in pass0,
  * blocks aren't used and we are scaning used part only */

    if (pos + NR_TO_READ < SB_JOURNAL_BLOCK (&g_sb) + JOURNAL_BLOCK_COUNT + 1)
	return 1; /* skip not data area of the filesystem (journal, reserved space) */

    return (!are_new_blocks_formatted(pos));
}

#if 0
void handle_leaf_in_pass1(struct buffer_head *bh, struct si ** saved_items)
{
    if (is_leaf_bad (bh)) {
        /* leaf is bad: directory item structure corrupted, or something else */
        /* mark_formatted_pointed_by_indirect (block);*/
        if (opt_verbose)
            reiserfs_warning ("\nhandle_leaf_in_pass1: bad leaf encountered: %lu\n", bh->b_blocknr);
        add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
        /* Save good items only to put them into tree at the end of this pass */
        for_all_good_items_in_node (save_item, saved_items, pointers_in_pass1, bh, SAVE);
        brelse (bh);
        return;
    }

    if (is_block_uninsertable (cur_pos)) {
        /* all leaves which are pointed to as unformatted nodes are marked in g_uninsertable bitmap */
        if (opt_verbose)
            reiserfs_warning ("\nhandle_leaf_in_pass1: leaf pointed to as unfm encountered: %lu\n", bh->b_blocknr);
        add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
        /* Rather than try to find UNP to this block we save its items and will put them
           into tree at the end of pass 1 */
        unmark_block_uninsertable(cur_pos);
        for_all_items_in_node (save_item, saved_items, bh, SAVE);
        brelse (bh);
	return;
    }

    if (tree_is_empty () == 1) {
        make_single_leaf_tree (bh);
        return;
    }

    /* if the leaf node can not be inserted into tree by pointer,		
       we postpone its insertion at the end of the pass 1 */
    try_to_insert_pointer_to_leaf (bh);
}
#endif

#if 0

/* bbh is pointer to read blocks */
/* count - number of read blocks or number of blocks to read */
static int read_some_blocks (reiserfs_filsys_t fs, struct buffer_head ** bbh,
			     int * count)
{
    __u32 temp = cur_pos % NR_TO_READ;

    if (temp != 0)
        die ("read_some_blocks: bad position occurs, bit must be 0");

    *count = 0;

    if (cur_pos >= SB_BLOCK_COUNT (fs))
        return OUT_OF_BITMAP;

    if (cur_pos + NR_TO_READ > SB_BLOCK_COUNT (fs)){
        *count = SB_BLOCK_COUNT (fs) - cur_pos;
        return LAST_BITS;
    }

    if (fsck_stage (fs) == PASS0 && some_blocks_rule_for_pass0(cur_pos) ||
        fsck_stage (fs) == PASS1 && some_blocks_rule_for_pass1(cur_pos) )
        return NOT_USED;

    *bbh = bread (fs->s_dev, cur_pos / NR_TO_READ, fs->s_blocksize * NR_TO_READ);

    if (*bbh == 0)
        return BAD_BLOCK;

    *count = NR_TO_READ;
    return 0;
}

/* bbh is pointer to read blocks */
/* count - number of read blocks or number of blocks to read */
int read_one_block(struct buffer_head ** bbh, int * count)
{
    *count = 0;

    if (cur_pos >= SB_BLOCK_COUNT (&g_sb))
        return OUT_OF_BITMAP;

    if (fsck_stage == PASS0 && one_block_rule_for_pass0(cur_pos) ||
        fsck_stage == PASS1 && one_block_rule_for_pass1(cur_pos) )
        return NOT_USED;

    *bbh = bread (g_sb.s_dev, cur_pos, g_sb.s_blocksize);

    if (*bbh == 0)
        return BAD_BLOCK;

    *count = 1;
    return 0;
}




/* after pass0 we have :
 *      all used leaves marked as formatted
 *      all used blocks which pointed to as unfm blocks marked as unformatted
 *      all bad blocks are marked as used
 *      all poiters to unused part are zeroed
 *
 * mark_bad_pointed_blocks_as_uninsertable() before pass1 */

/**/
static void go_through (reiserfs_filsys_t fs, struct si ** saved_items,
			__u32 blocks_to_read)
{
    __u32 handled_blocks = 0;
    int count;
    int blocks_to_read_by_1 = 0; /* number of blocks to read by 1 */
    int excep = 0;
    struct buffer_head * bbh, *bh;
    char * data;

    if ( opt_fsck == 0 )
        if (fsck_stage (fs) == PASS0)
            fprintf (stderr, "\nPass 0 - ");
        else
            fprintf (stderr, "\nPass 1 - ");
    fflush(stdout);

    /* position?? */
    cur_pos = -1;

    while (1) {
        int k;

        if (blocks_to_read_by_1 == 0) {
            switch (excep){     /* read_some_blocks returns one of these exeptions */
            case LAST_BITS:                                             /* we should read by 1 block */
            case BAD_BLOCK:     blocks_to_read_by_1 = count;    break;  /* we should read by 1 block */
            case NOT_USED:      cur_pos += 8;                   break;  /* we should read next 8 blocks */
            case 0:             cur_pos ++;                     break;  /* we should read next 8 blocks */
            }
        }else{
            /* we should read_one_block again */
            blocks_to_read_by_1 --;
            cur_pos++;
        }

        if (!blocks_to_read_by_1) {
            excep = read_some_blocks(&bbh, &count);
        }else{
            excep = read_one_block(&bbh, &count);
            if (excep == BAD_BLOCK) {
                mark_block_used(cur_pos);
                reiserfs_warning("\nbad block found (%u)\n", cur_pos);
                if ( opt_fsck == 0 )
                    print_how_far (&handled_blocks, blocks_to_read);
            }
        }

        switch (excep){
            case OUT_OF_BITMAP:  return;
            case NOT_USED:
            case LAST_BITS:
            case BAD_BLOCK:      continue;
        }

        for (k = 0; k < count; k ++) {
            int what_node;

            if (k) /* if we read more then 1 block we should correct position */
                cur_pos ++;

            if ( fsck_stage == PASS0 && one_block_rule_for_pass0(cur_pos) ||
                 fsck_stage == PASS1 && one_block_rule_for_pass1(cur_pos) )
                continue;       /* current block is free and we scan only used part */

            if ( opt_fsck == 0 )
                print_how_far (&handled_blocks, blocks_to_read);

            data = bbh->b_data + k * fs->s_blocksize;

	    what_node = who_is_this (data, fs->s_blocksize);
	    if ( what_node != THE_LEAF )
		continue;

            /* leaf node is selected */
            bh = make_buffer (fs->s_dev, cur_pos, fs->s_blocksize, data);

            if (fsck_stage == PASS1) {
                /* bbh is a formatted */
                handle_leaf_in_pass1(bh, saved_items);
                continue;
            }

            if (fsck_stage == PASS0) {
		mark_block_formatted(cur_pos);
		g_blocks_to_read_in_pass1++;
//		for_items (0, 0, bh, pointers_in_pass0, 0);
		for_all_good_items_in_node (0, 0, pointers_in_pass0, bh, 0);
		brelse (bh);
            }
        }/* for  */
        bforget (bbh);
    }/* while */
}

#endif


/*
 * go_through changed a bit 
 */

typedef int (should_check_f) (reiserfs_filsys_t, unsigned long);
typedef void (handle_leaf_f) (reiserfs_filsys_t, struct buffer_head *);




/* in pass 1 we read all blocks marked as formatted (leaves not pointed by
   indirect items) */
static int should_read_in_pass1 (reiserfs_filsys_t fs, unsigned long block)
{
    if (reiserfs_bitmap_test_bit (new_bitmap (fs), block))
	die ("should_read_in_pass1: block %lu is in tree already");

    return reiserfs_bitmap_test_bit (found_leaves (fs), block);
}


/* if any of blocks in the interval is to be read - we read whole interval */
static int should_check_any (reiserfs_filsys_t fs, unsigned long from, int count,
			     should_check_f should_check)
{
    unsigned long i;

    for (i = from; i < from + count; i ++)
	if (should_check (fs, i))
	    return 1;
    return 0;
}


/* in pass 1 we try to put perfectly correct leaves into tree, so these
   is_bad_*_pass1 handlers have to check as much as possible.  */
static int is_bad_sd_pass1 ()
{
}

static int is_bad_indirect_pass1 ()
{
}


static int is_bad_direct_pass1 ()
{
}

static int is_bad_direntry_pass1 ()
{
}


static void handle_leaf_in_pass1 (reiserfs_filsys_t fs, struct buffer_head * bh, struct si ** saved_items)
{
    if (is_leaf_bad (bh)) {
        /* leaf is bad: directory item structure corrupted, or something else */
        /* mark_formatted_pointed_by_indirect (block);*/
        if (opt_verbose)
            reiserfs_warning ("\nhandle_leaf_in_pass1: bad leaf encountered: %lu\n", bh->b_blocknr);
        add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
        /* Save good items only to put them into tree at the end of this pass */
        for_all_good_items_in_node (save_item, saved_items, pointers_in_pass1, bh, SAVE);
        brelse (bh);
        return;
    }

    if (is_block_uninsertable (cur_pos)) {
        /* all leaves which are pointed to as unformatted nodes are marked in g_uninsertable bitmap */
        if (opt_verbose)
            reiserfs_warning ("\nhandle_leaf_in_pass1: leaf pointed to as unfm encountered: %lu\n", bh->b_blocknr);
        add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
        /* Rather than try to find UNP to this block we save its items and will put them
           into tree at the end of pass 1 */
        unmark_block_uninsertable(cur_pos);
        for_all_items_in_node (save_item, saved_items, bh, SAVE);
        brelse (bh);
	return;
    }

    if (tree_is_empty () == 1) {
        make_single_leaf_tree (bh);
        return;
    }

    /* if the leaf node can not be inserted into tree by pointer,		
       we postpone its insertion at the end of the pass 1 */
    try_to_insert_pointer_to_leaf (bh);
}


/* 'should_check' says whether the node is to be read, checked with
   'handle_leaf' if it is a leaf */
static void go_through_2 (reiserfs_filsys_t fs, struct si ** saved_items,
			  should_check_f should_check, handle_leaf_f handle_leaf)
{
    struct buffer_head * bbh, * bh;
    int nr_to_read = NR_TO_READ;
    int i, j;
    char * data;
    int what_node;

    for (i = 0; i < SB_BLOCK_COUNT (fs); i += nr_to_read) {
	if (should_check_any (fs, i, nr_to_read, should_check)) {
	    /* at least one of nr_to_read blocks is to be checked */
	    bbh = bread (fs->s_dev, i / nr_to_read, fs->s_blocksize * nr_to_read);
	    if (bbh) {
		for (j = 0; j < nr_to_read; j ++) {
		    if (should_check (fs, i + j)) {
			data = bbh->b_data + j * fs->s_blocksize;
			
			what_node = who_is_this (data, fs->s_blocksize);
			if ( what_node != THE_LEAF )
			    continue;
			
			/* the node looks like a leaf, but it still can be
                           not perfect */
			bh = make_buffer (fs->s_dev, i + j, fs->s_blocksize, data);
			handle_leaf (fs, bh);
			brelse (bh);
		    }
		}
		bforget (bbh);
	    } else {
		/* bread failed */
		if (nr_to_read != 1) {
		    /* we tryied to read bunch of blocks. Try to read them by one */
		    i -= nr_to_read;
		    nr_to_read = 1;
		}
	    }
	}

	if (nr_to_read == 1 && ((i + 1) % NR_TO_READ) == 0) {
	    /* we have read NR_TO_READ blocks one at time, switch back to
               reading NR_TO_READ blocks at time */
	    i -= (NR_TO_READ - 1);
	    nr_to_read = NR_TO_READ;
	}
    }
    if (i != SB_BLOCK_COUNT (fs))
	die ("go_through_2: block counting failed");
}



/* in pass 0 we read all blocks marked used in on-disk bitmap but blocks
   prealocated on mkreiserfs time */
static int should_read_in_pass0 (reiserfs_filsys_t fs, unsigned long block)
{
    if (not_data_block (fs, block))
	return 0;

    return reiserfs_bitmap_test_bit (disk_bitmap (fs), block);
}


/* pass 0: mark unformatted in the bitmap of nodes pointed to by indirect
   items. If unformatted node pointer is too big - it is skipped. If it points
   to a block which is not marked used in on-disk bitmap - zero that slot
   (fixme: we could instead mark it used) */
static void handle_unfm_ptr_pass0 (reiserfs_filsys_t fs, 
				   struct buffer_head * bh,
				   __u32 * unfm_ptr)
{
    __u32 block;

    block = le32_to_cpu (*unfm_ptr);
    if (!block)
	return;
    if (block >= SB_BLOCK_COUNT (fs))
	return;
    if (!reiserfs_bitmap_test_bit (disk_bitmap (fs), block)) {
	*unfm_ptr = 0;
	mark_buffer_dirty (bh, 1);
	return;
    }
    reiserfs_bitmap_set_bit (unformatted_bitmap (fs), block);
}


/* for every slot in indirect items call handle_unfm_ptr */
static void for_every_unfm_ptr (reiserfs_filsys_t fs,
				struct buffer_head * bh, 
				void handle_unfm_ptr (reiserfs_filsys_t fs,
						      struct buffer_head * bh,
						      __u32 * unfm_ptr))
{
    int i, j;
    struct item_head * ih;
    __u32 * indirect;

    ih = B_N_PITEM_HEAD (bh, 0);
    for (i = 0; i < node_item_number (bh); i ++, ih ++) {
	if (!is_indirect_ih (ih))
	    continue;
	indirect = (__u32 *)B_I_PITEM (bh, ih);
	for (j = 0; j < I_UNFM_NUM (ih); j ++, indirect ++)
	    handle_unfm_ptr (fs, indirect);
    }
	    
}


/* set bit in bitmap of leaves, check indirect items and mark pointed blocks
   in bitmap of unformatted */
static void handle_leaf_in_pass0 (reiserfs_filsys_t fs, struct buffer_head * bh)
{
    reiserfs_bitmap_set_bit (found_leaves (fs), bh->b_blocknr);
    
    for_every_unfm_ptr (fs, bh, handle_unfm_ptr_pass0);
}



/* read all blocks marked in on-disk bitmap as used but journal area */
void pass0 (reiserfs_filsys_t fs)
{
    fsck_stage (fs) = PASS0;

    /* read blocks if should_read_in_pass0 returns 1 and call
       handle_leaf_in_pass0 it that block looks like a leaf */
    go_through_2 (fs, 0, /* save items */
		  should_read_in_pass0, handle_leaf_in_pass0);
    
    if (fsck_mode (fs) != FSCK_REBUILD)
	die ("who else needs pass 0 but rebuild?");

    /* for every bit in bitmap */
    for (i = 0; i < SB_BLOCK_COUNT (fs); i ++) {
	if (reiserfs_bitmap_test_bit (found_leaves (fs), i) &&
	    reiserfs_bitmap_test_bit (found_unfm_ptrs (fs), i))
	    /* block looks like leaf and pointed by some indirect item */
	    reiserfs_bitmap_clear_bit (not_for_insertion (fs), i);

	if (!not_data_block (i) &&
	    reiserfs_bitmap_test_bit (disk_bitmap (fs), i) &&
	    !reiserfs_bitmap_test_bit (found_leaves (fs), i) &&
	    !reiserfs_bitmap_test_bit (found_unfm_ptrs (fs), i)) {
	    /* this is internal, mark it free */
	    reiserfs_bitmap_clear_bit (disk_bitmap (fs), i);
	    set_free_blocks (fs->s_rs, rs_free_blocks (fs->s_rs) + 1);
	}
    }

    /* reset unformatted bitmap */
    reiserfs_bitmap_zero (found_unfm_ptrs (fs));
}


/* pass1 */
void build_the_tree (reiserfs_filsys_t fs)
{
    __u32 handled_blocks = 0;
    struct si * saved_items = 0;

    fsck_stage (fs) = PASS1;

    go_through_2 (fs, &saved_items, should_read_in_pass1, handle_leaf_in_pass1);


    /*  put saved items into tree. These items were in leaves, those could not
	be inserted into tree because some indirect items point to those
	leaves. Rather than lookup for corresponding unfm pointers in the
	tree, we save items of those leaves and put them into tree separately */

    if (SB_ROOT_BLOCK(&g_sb) == -1)
        die("\n\n No one good block was found, try --scan-whole-partition");

    if ( opt_fsck == 0 )
	fprintf (stderr, "\nPass 1a - ");
	
    fsck_stage (fs) = PASS1a;
    put_saved_items_into_tree (saved_items);

    if ( opt_fsck == 0 )
	fprintf (stderr, "done\n\n");

    /* end of pass 1 */
    /* this works only if -c specified  */
    reiserfsck_check_pass1 ();

    if (opt_stop_point == STOP_AFTER_PASS1)
	return;
	
    /* pass 2 */
    fsck_stage = PASS2;
    take_bad_blocks_put_into_tree ();

    reiserfsck_check_pass1 ();
}

#if 0
/* reads the device by set of 8 blocks, takes leaves and tries to
   insert them into tree */
void build_the_tree_old (void)
{
    int i, j, k, max_k;
    struct buffer_head * bbh, * bh;
    __u32 handled_blocks = 0;
    struct si * saved_items = 0;
    char * data;

    if ( opt_fsck == 0 )
        fprintf (stderr, "Pass 1 - ");

    fflush(stdout);
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++)
        for (j = 0; j < g_sb.s_blocksize; j ++) {
                /* make sure, that we are not out of the device */
            if (i * g_sb.s_blocksize * 8 + j * 8 >= SB_BLOCK_COUNT (&g_sb))
                goto out_of_bitmap;

            max_k = (i * g_sb.s_blocksize * 8 + j * 8 + 8 > SB_BLOCK_COUNT (&g_sb)) ?
                SB_BLOCK_COUNT (&g_sb) - (i * g_sb.s_blocksize * 8 + j * 8) : 8;

            if (SB_AP_BITMAP (&g_sb)[i]->b_data[j] == 0 && opt_what_to_scan == SCAN_USED_PART)
                    continue; /* all blocks are free */

	    if (opt_fsck_mode == FSCK_FAST_REBUILD &&
	        (SB_AP_BITMAP (&g_sb)[i]->b_data[j] & (~g_new_bitmap[i][j])) == 0)
                continue; /* --fast-rebuild & !(any block was_used & not in tree yet) */
	    	
	    bbh = bread (g_sb.s_dev, i * g_sb.s_blocksize + j, g_sb.s_blocksize * 8);
//	            printf("\n bread %d", (i * g_sb.s_blocksize + j) * 8 );
            for (k = 0; k < max_k; k ++) {
                unsigned long block;
                int what_node;

                if ((SB_AP_BITMAP (&g_sb)[i]->b_data[j] & (1 << k)) == 0 && opt_what_to_scan == SCAN_USED_PART)
                    continue; /* k-th block is free */

                block = i * g_sb.s_blocksize * 8 + j * 8 + k;
		
                if (opt_fsck_mode == FSCK_FAST_REBUILD &&
                        (SB_AP_BITMAP (&g_sb)[i]->b_data[j] & (1 << k)) == 1 && is_block_used(block))
                        /* this block is in tree already(we didn't remove it from the tree
                                during cure_internal_tree()) */
                    continue;

		    		
                if ( opt_fsck == 0 )
                    print_how_far (&handled_blocks, g_blocks_to_read_in_pass0);

                if (not_data_block (&g_sb, block)) {
                        /* skip not data area of the filesystem (journal, bitmaps, reserved space) */
                    continue;
                }


                if (bbh == 0)
                {
                    bh = bread (g_sb.s_dev, block, g_sb.s_blocksize);
                    if (bh == 0)
                    {
                        mark_block_used(block);
                        continue;
                    }
                    data = bh->b_data;
                }else
                    data = bbh->b_data + k * g_sb.s_blocksize;

                what_node = who_is_this (data, g_sb.s_blocksize);

                switch (what_node) {
                case THE_LEAF:
                case THE_INTERNAL:
                        break;
                default:
                        //
                        continue;
                }

                if (is_internal_block_head(data) == 1)
                    //if (is_internal_block_head1(bbh->b_data + k * g_sb.s_blocksize) == 1 && opt_fsck_mode == FSCK_REBUILD)
                {
                    if (opt_fsck_mode == FSCK_REBUILD)
                    {
                        if (!is_block_used (block))
                            reiserfs_free_internal_block (&g_sb, block);
    		        else
			    /* block is used in new tree already. There was an indirect item, pointing to it.
			    We keep information about it for check only */
			    /*mark_formatted_pointed_by_indirect (block)*/
                        ;
                    }
                    continue;
                }

                    /* leaf node is selected */
                bh = make_buffer (g_sb.s_dev, block, g_sb.s_blocksize, data);
	            /* */
                if (opt_fsck_mode == FSCK_FIND_ITEM) {
                    find_a_key (&key_to_find, offset_to_find, bh);
                    brelse (bh);
                    continue;
    		}

                if (is_block_used (block)) {
                        /* block is used in new tree already. There was an indirect item, pointing to it.
                        We keep information about it for check only */
                        /*  mark_formatted_pointed_by_indirect (block);*/
                    add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
                        /* Rather than try to find UNP to this block we save its items and will put them
		        into tree at the end of pass 1 */
                    for_all_items_in_node (save_item, &saved_items, bh, SAVE);
                    brelse (bh);
		    continue;
                }

                if (is_leaf_bad (bh)) {
                        /* leaf is bad: directory item structure corrupted, or something else */
                        /* mark_formatted_pointed_by_indirect (block);*/
                    if (opt_verbose)
            		reiserfs_warning ("\nbuild_the_tree: bad leaf encountered: %lu\n", bh->b_blocknr);
		    add_event (LEAVES_USED_BY_INDIRECT_ITEMS);
                        /* Save good items only to put them into tree at the end of this pass */
                    for_all_items_in_node (save_item, &saved_items, bh, SAVE);
                    brelse (bh);
                    continue;
    		}

                if (tree_is_empty () == 1) {
                    make_single_leaf_tree (bh);
                    continue;
                }

        	    /* if the leaf node can not be inserted into tree by pointer,		
		       we postpone its insertion at the end of the pass 1 */
                try_to_insert_pointer_to_leaf (bh);
		
                /*&&&&&&&&&&&&&&&&*/
                //reiserfsck_check_pass1 ();
                /*&&&&&&&&&&&&&&&&*/
            }

	    bforget (bbh);

	}

 out_of_bitmap:

/*
    check_bitmaps (&g_sb);

    printf ("bitmap scanning completed: allocated %d, freed %d *really freed %d, new internals %d, found internals %d (freed %d)\n",
	    g_allocated, g_freed, g_really_freed, g_new_internals, g_found_internals, g_freed_internals);fflush (stdout);
*/


    if (opt_fsck_mode == FSCK_FIND_ITEM)
        return;

    /* this checks what has been built (if -c option is set) */
/*    reiserfsck_check_pass1 ();*/

    /* put saved items into tree. These items were in leaves, those
       could not be inserted into tree because some indirect items point
       to those leaves. Rather than lookup for corresponding unfm
       pointers in the tree, we save items of those leaves and put them
       into tree separately */
    if ( opt_fsck == 0 )
	printf ("\nPass 1a - ");
	
    put_saved_items_into_tree (saved_items);

    if ( opt_fsck == 0 )
	printf ("done\n");

    /* end of pass 1 */
    if ( opt_fsck == 0 )
	printf ("\n");

    /* this works only if -c specified  */

    if (SB_ROOT_BLOCK(&g_sb) == -1)
        die("\n\n No one item was found, try --scan-whole-partition");

    reiserfsck_check_pass1 ();


    if (opt_stop_point == STOP_AFTER_PASS1)
	return;
	
    /* pass 2 */
    take_bad_blocks_put_into_tree ();

    reiserfsck_check_pass1 ();
}
#endif




/*
for fast-rebuild all internals must be marked as formatted in pass0.
(we should not read this block again in fast-pass if it is not formatted)

Probably we should read by 8 block in pass_through_tree().

Bad leaves we should save here to insert later item-by-item.

Duplicate unfm pointers (or pointers to leaf or internals) found in fast-pass
should NOT be cured right here - if there is a wrong block connected to the tree
(for examlpe, wrong leaf) too meny items will be postponed and saved in saved_items.

If leaf is OK, but marked as unfm also, than the leaf is most
likely OK, but pointer is not. Or just copy this leaf into free block.

If internal is OK, but marked as unfm also, it should be copied into
free block. (There can be the situation when for the 1st look it is
good internal but all pointers are deleted later)

If leaf itself is bad it should be saved into saved_items right here.


*/

/* main FAST-REBUILD function */
/* Look through inesrtnal tree and remove non-consistent parts from the tree */
/* If block is formatted and is in tree after this function and we found pointer to this block
   it should be considered as a bad pointer and a good leaf/internal */
void recover_internal_tree (struct super_block * s)
{
    return;

    /* control bitmap is used to keep all blocks we should not put into tree again */
    /* used bitmap is used to keep all inserted blocks. The same as control bitmap plus unfm blocks */

    /* look through tree structure and remove non consistent parts from the tree */
    check_internal_structure(s);
    /* trying to insert good leaves as a whole and bad leaves item-by-item */
    build_the_tree();
}

/* sb_not_found == uread_super_block (&g_sb) */
/* this finds SB on usual places and corrects data if needed */
/* if SB is not found we prompt to user to create SB of proper version again */
void rebuild_sb (int sb_not_found)
{
    int version;
    struct buffer_head * bh;
    struct reiserfs_super_block * rs;
    __u32 blocks;

    g_sb.s_blocksize = 4096;
    if(sb_not_found)
    {   /*menu for user*/
        printf("\nwhat is version of ReiserFS you use[1-4]\n");
        printf("\t(1)   3.6.x\n");
        printf("\t(2) >=3.5.9\n");
        printf("\t(3) < 3.5.9 converted to new format\n");
        printf("\t(4) < 3.5.9\n");
        printf("\t(X)   exit\n");
	    if (scanf ("%d", &(version)) != 1)
    	    die ("rebuild_sb: specify a version through stdin");
        if (version < 1 || version > 4)
    	    die ("rebuild_sb: wrong version");
    	
        /* read different blocks according to the version */
        switch(version){
        case 1:
        case 2:
            bh = bread (g_sb.s_dev, (REISERFS_DISK_OFFSET_IN_BYTES / g_sb.s_blocksize), g_sb.s_blocksize);
            break;
        case 3:
        case 4:
            bh = bread (g_sb.s_dev, (2), g_sb.s_blocksize);
            break;
        default:
            exit(0);
        }
        if (!bh)
            die ("rebuild_sb: can't bread");
        rs = (struct reiserfs_super_block *)bh->b_data;
        g_sb.s_rs = rs;
        SB_BUFFER_WITH_SB (&g_sb) = bh;
    }
    else
    {
        char * yes = 0;
        size_t n = 0;

        /* trying to find a SB */
        version = check_sb(&g_sb);
        printf("\nDo you want to remake your super block\n(say no if you use resizer)[Yes/no]:\n");
        if (getline (&yes, &n, stdin) != 4 || strcmp ("Yes\n", yes))
        	exit (0);
        rs = g_sb.s_rs;
    }

    /* set block number on the device and number of bitmap blocks needed to address all blocks */
    blocks = (count_blocks ("", g_sb.s_blocksize, g_sb.s_dev) / 8) * 8;
    set_block_count (rs, blocks);
    //rs->s_block_count = cpu_to_le32(blocks);

    set_bmap_nr (rs, (blocks + (g_sb.s_blocksize * 8 - 1)) / (g_sb.s_blocksize * 8));
    set_journal_size(rs, JOURNAL_BLOCK_COUNT);

    //rs->s_bmap_nr = cpu_to_le16( blocks / (g_sb.s_blocksize * 8) +
    //      ((blocks % (g_sb.s_blocksize * 8)) ? 1 : 0) );

    switch (version){
    case 1:
	/* super block v2 at 64k offset */
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISER2FS_SUPER_MAGIC_STRING,
		 strlen(REISER2FS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_start_must (g_sb.s_blocksize));
	set_version (rs, REISERFS_VERSION_2);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE) / sizeof(__u32) / 2 * 2);
	break;

    case 2:
	/* super block v1 at 64k offset */
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISERFS_SUPER_MAGIC_STRING,
		 strlen(REISERFS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_start_must (g_sb.s_blocksize));
	set_version (rs, REISERFS_VERSION_1);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE_V1) / sizeof(__u32) / 2 * 2);
	break;

    case 3:
	/* super block v2 at 8k offset */
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISER2FS_SUPER_MAGIC_STRING,
		 strlen(REISER2FS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_old_start_must (rs));
	set_version (rs, REISERFS_VERSION_2);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE) / sizeof(__u32) / 2 * 2);
	break;

    case 4:
	/* super block v1 at 8k offset */
	set_blocksize (rs, g_sb.s_blocksize);
	strncpy (rs->s_v1.s_magic, REISERFS_SUPER_MAGIC_STRING,
		 strlen(REISERFS_SUPER_MAGIC_STRING));
	set_journal_start (rs, get_journal_old_start_must (rs));
	set_version (rs, REISERFS_VERSION_1);
	set_objectid_map_max_size (rs, (g_sb.s_blocksize - SB_SIZE_V1) / sizeof(__u32) / 2 * 2);
	break;
    }

    mark_buffer_dirty(SB_BUFFER_WITH_SB (&g_sb), 1);
    reiserfs_warning("\nDo not forget to run reiserfsck --rebuild-tree\n\n");
}






