/*
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */

#include "fsck.h"


/* size of 1 piece of map */
#define MAP_SIZE   1024 /* must be n * 2 * sizeof(__u32) */
#define MAX_MAP_SIZE 1 /* % of available memory? */



/* increase area by MAP_SIZE bytes */
static void grow_id_map (struct id_map * map)
{
    map->m_begin = expandmem (map->m_begin, map->m_count * MAP_SIZE, MAP_SIZE);
    map->m_count++;
}


static void try_to_shrink_id_map (struct id_map * map)
{
    if (map->m_used * sizeof(__u32) <= (map->m_count - 1) * MAP_SIZE) {
	map->m_begin = expandmem (map->m_begin, map->m_count * MAP_SIZE,
				  -MAP_SIZE);
        map->m_count--;
    }
}


/* ubin_search_id is used to find id in the map (or proper place to
   insert the new id).  if smth goes wrong or ubin_search_id stops
   working properly check_id_search_result should help to find raised
   problems */
static int check_id_search_result(struct id_map * map, int res, __u32 pos,
				  __u32 id)
{
    if (res != 1 && res != 0)
        die("check_id_search_result: get wrong result from ubin_search (%d)", res);

    if (res == 1 && *(map->m_begin + pos) != id)
        die("check_id_search_result: wrong id found %u %u", id, *(map->m_begin + pos));

    if (res == 1)
    {
        if (pos > map->m_used)
            die("check_id_search_result: get bad position (%u), used %u", 
		pos, map->m_used);
        if (pos >= 0 && pos <= map->m_used && *(map->m_begin + pos - 1) >= id)
            die("check_id_search_result: previous id (%u) not less than (%u)",
		*(map->m_begin + pos - 1), id);
        if (pos >= 0 && pos < map->m_used && *(map->m_begin + pos) < id)
            die("check_id_search_result: found id (%u) not much than (%u)",
		*(map->m_begin + pos), id);
    }
}


static int comp_ids (__u32 * id1, __u32 * id2)
{
    if ( *id1 < *id2 )
        return -1;
    if ( *id1 > *id2 )
        return 1;
    return 0;
}



void init_id_map (reiserfs_filsys_t fs)
{
    proper_id_map(fs).m_begin = NULL;
    proper_id_map(fs).m_used = 0;
    proper_id_map(fs).m_count = 0;
    mark_id_as_used (fs, 1);
}


/* free whole map */
void free_id_map (reiserfs_filsys_t fs)
{
    freemem (proper_id_map(fs).m_begin);
    proper_id_map(fs).m_begin = 0;
    proper_id_map(fs).m_used = 0;
    proper_id_map(fs).m_count = 0;
}


/* return 1 if id is marked used, 0 otherwise */
int is_id_used (reiserfs_filsys fs, __u32 id, __u32 * ppos)
{
    int res;
    struct id_map * map;
    
    map = &proper_id_map (fs);

    *ppos = 0;

    if (map->m_begin == NULL)
        return 0;

    /* smth exists in the map, find proper place to insert or this id */
    res = ubin_search (&id, map->m_begin, map->m_used, ppos, comp_ids);
#if 1
    check_id_search_result (map, res, *ppos, id);
#endif
    /* *ppos is position in objectid map of the element which is equal id
       or position of an element which is smallest and greater than id */
    if (!res)
	/* id is not found in the map. if returned position is odd -
	   id is marked used */
	return (*ppos % 2);

    /* if returned position is odd - id is marked free */
    return !(*ppos % 2);
}


/* returns 1 if error occurs, 0 otherwise */
int mark_id_as_used (reiserfs_filsys_t fs, __u32 id)
{
    __u32 pos;
    __u32 * objectid_map;
    struct id_map * map;
    
    /* check whether id is used and get place if used or place to insert if not */
    if (is_id_used (fs, id, &pos) == 1)
        return 1;

    map = &proper_id_map (fs);
    if (pos % 2 == 0){
        /* id not found in the map. why? is_id_used() knows */

        if (map->m_begin == NULL)
	    /* map is empty */
            grow_id_map (map);

        /* id + 1 is used, change id + 1 to id and exit */
        if ( id + 1 == le32_to_cpu (map->m_begin[pos]) ) {
	    /* we can mark id as used w/o expanding of id map */
	    map->m_begin[pos] = cpu_to_le32 (id);
	    return 0;
        }

        if (map->m_count * MAP_SIZE == map->m_used * sizeof(__u32))
	    /* fixme: do not grow too much */
            grow_id_map (map);

        if (map->m_used - pos > 0)
            memmove (map->m_begin + pos + 2, map->m_begin + pos, (map->m_used - pos) * sizeof (__u32));

        map->m_used += 2;
        map->m_begin[pos] = cpu_to_le32 (id);
        map->m_begin[pos+1] = cpu_to_le32 (id + 1);
        return 0;
    }

    /* id found in the map. pos is odd position () */
    map->m_begin[pos] = cpu_to_le32 (id + 1);
    
    /* if end id of current interval == start id of next interval we
       eliminated a sequence of unused objectids */
    if (pos + 1 < id_map.m_used &&
	map->m_begin[pos + 1] == map->m_begin[pos]) {
	memmove (map->m_begin + pos, map->m_begin + pos + 2, (map->m_used - pos - 2) * sizeof (__u32));
	map->m_used -= 2;
	try_to_shrink_id_map (map);
    }
    return 0;
}


__u32 get_free_id (reiserfs_filsys_t fs)
{
    struct id_map * map;

    map = &proper_id_map (fs);

    /* If map is not NULL return the second element (first position in
       the map).  This allocates the first unused objectid. That is to
       say, the first entry on the objectid map is the first unused
       objectid. */
    if (map->m_begin == NULL) {
	fprintf (stderr, "get_free_id: hmm, 1 is allocated as objectid\n");
	return 1;
    }
    return (le32_to_cpu (map->m_begin[1]));
}


#define objectid_map(fs) (((char *)((fs)->s_rs) + sb_size (fs)))


/* returns 0 if on-disk objectid map matches to the correct one, 1
   otherwise */
int compare_id_maps (reiserfs_filsys_t fs)
{
    struct id_map * map;
    int disk_size;
  
    map = &proper_id_map (fs);
    
    disk_size = rs_objectid_map_size (fs->s_rs);
    if (disk_size != map->m_used ||
	memcmp (objectid_map (fs), map->m_begin, sizeof(__u32) * disk_size)) {
	fprintf (stderr, "Objectid maps mismatch\n");
	return 1;
    }
    if (fs->fu_job->verbose)
	fprintf (stderr, "Objectid maps match\n");

    return 0;
}


/* copy objectid map into buffer containing super block */
void correct_objectid_map (reiserfs_filsys_t fs)
{
    struct id_map * map;
    int size, disk_max;
 
    map = &proper_id_map (fs);
    
    size = map->m_used;
    disk_max = rs_objectid_map_max_size (fs->s_rs);
    if (disk_max < size) {
	size = disk_max;
    } else {
	memset (objectid_map (fs) + size, 0, (disk_max - size) * sizeof (__u32));
    }
    
    memcpy (objectid_map (fs), map->m_begin, size * sizeof (__u32));
    set_objectid_map_size (fs, size);
    mark_buffer_dirty (SB_BUFFER_WITH_SB (fs), 1);

    if (fs->fu_job->verbose)
	fprintf (stderr, "Objectid map corrected\n");
}




/* print the map of objectids */
void print_objectid_list ()
{
    int i;
    printf ("\n control id map: all %d, used:%d", id_map.m_count * MAP_SIZE, id_map.m_used);

    for (i = 0; i < id_map.m_used; i += 2)
	    printf ("\n[%u-%u]", id_map.m_begin[i], id_map.m_begin[i + 1] - 1);
}

/* print on-disk map of objectids */
void print_disk_objectid_list ()
{
    int i;
    __u32 * objectid_map = (__u32 *)((char *)SB_DISK_SUPER_BLOCK (&g_sb) + (sb_size(&g_sb)));
    printf ("\n on-disk id map. used:%lu", SB_OBJECTID_MAP_SIZE(&g_sb));

    for (i = 0; i < SB_OBJECTID_MAP_SIZE(&g_sb); i += 2)
	    printf ("\n[%u-%u]", objectid_map[i], objectid_map[i + 1] - 1);
}

